import openai
import os
from dotenv import load_dotenv
from pydantic import BaseModel
from typing import List, Optional

load_dotenv()

import tanuki
from utils import scrape_url


openai.api_key = os.getenv("OPENAI_API_KEY")


class Quote(BaseModel):
    text: str
    author: str
    tags: List[str] = []


@tanuki.patch
def extract_quote(content: str) -> Optional[Quote]:
    """
    Examine the content string and extract the quote details for the text, author, and tags.
    """


@tanuki.align
def align_extract_quote() -> None:
    print("Aligning...")
    quote = "\nIt takes courage to grow up and become who you really are.\nby E.E. Cummings\n(about)\n\n\n            Tags:\n            \ncourage\n\n"
    assert extract_quote(quote) == Quote(
        text="It takes courage to grow up and become who you really are.",
        author="E.E. Cummings",
        tags=["courage"],
    )


if __name__ == '__main__':

    # Align the function
    align_extract_quote()

    # Web scrape the url and extract the list of quotes
    url = "https://quotes.toscrape.com/page/1/"
    contents = scrape_url(url=url, class_name="quote")

    # Process the quote blocks using Tanuki (only sampling a couple for demo purposes)
    quotes = []
    for content in contents[0:2]:
        c = content.replace('“', '')
        c = c.replace('”', '')
        quotes.append(extract_quote(c))
    print(quotes)



from bs4 import BeautifulSoup
import os
import requests
from typing import List, Optional


# Update request header if user agent is provided
USER_AGENT = os.getenv("USER_AGENT", None)

headers = {}
if USER_AGENT:
    headers['User-Agent'] = USER_AGENT


def scrape_url(url: str, class_name: Optional[str] = None) -> List[str]:
    """Scrape the url and return the raw content(s) of all entities with a given
    div class name."""
    print(f"Scraping {url=} for {class_name=} entities...")

    response = requests.get(url, headers=headers)
    contents = []

    # Parse html content if response is successful
    if response.status_code == 200:
        soup = BeautifulSoup(response.text, 'html.parser')

        # Loop through the matched elements to get contents if a class name is provided
        if class_name:
            entities = soup.find_all('div', class_=class_name)
            for entity in entities:
                content = entity.text
                content = content.lstrip().rstrip()
                contents.append(content)

        # Otherwise, just get the text content
        else:
            content = soup.text
            content = content.lstrip().rstrip() 
            contents.append(content)

    # Print error if response is unsuccessful
    else:
        print(response.status_code)
        print(response.text)

    return contents



from collections import deque, defaultdict
from dataclasses import dataclass
from typing import List, Dict, Tuple, Union, Optional, Sequence, Iterable, Mapping, ChainMap, Set, get_args, FrozenSet, \
    DefaultDict, Counter, OrderedDict

import pytest
from pydantic import BaseModel

from tanuki.validator import Validator


def test_instantiate_pydantic():
    class Person(BaseModel):
        name: str
        age: int
        height: float
        is_cool: bool

        def __eq__(self, other):
            return self.model_dump() == other.model_dump()

        def __hash__(self):
            return hash(str(self.model_dump()))

    validator = Validator()
    person = {
        "name": "John",
        "age": 20,
        "height": 1.8,
        "is_cool": True
    }
    person_obj = validator.instantiate(person, Person)
    assert isinstance(person_obj, Person)

    # Nested data classes or Pydantic models.
    @dataclass
    class InnerDataClass:
        y: int

    @dataclass
    class OuterDataClass:
        x: InnerDataClass

    data = {"x": {"y": 10}}
    assert validator.instantiate(data, OuterDataClass) == OuterDataClass(InnerDataClass(10))

def test_instantiate_dataclass():
    validator = Validator()
    @dataclass
    class Person:
        name: str
        age: int
        height: float
        is_cool: bool

    person = {
        "name": "John",
        "age": 20,
        "height": 1.8,
        "is_cool": True
    }
    person_obj = validator.instantiate(person, Person)
    assert isinstance(person_obj, Person)

def test_primitives():
    """
    Test that primitives and empty objects are instantiated correctly.
    """
    validator = Validator()
    assert validator.instantiate("1", int) == 1
    assert validator.instantiate("1.0", float) == 1.0
    assert validator.instantiate("true", bool) == True
    assert isinstance(validator.instantiate("1", float), float)
    assert isinstance(validator.instantiate("1.0", int), int)
    assert validator.instantiate("1", str) != 1
    assert validator.instantiate("1.0", str) != 1.0
    assert validator.instantiate("true", str) != True
    assert validator.instantiate({}, dict) == {}
    assert validator.instantiate([], list) == []
    assert validator.instantiate((), tuple) == ()
    assert validator.instantiate(set(), frozenset) == set()
    assert validator.instantiate((), frozenset) == ()
    assert validator.instantiate((), set) == ()

def test_generics():
    """
    Test that generics are instantiated correctly.
    """
    validator = Validator()

    # Lists, dictionaries, tuples, sets, and frozensets.
    assert isinstance(validator.instantiate([1, 2, 3], List[int]), List)
    assert validator.instantiate({"a": 1, "b": 2}, Dict[str, int]) == {"a": 1, "b": 2}
    assert validator.instantiate((1, "a", 2.0), Tuple[int, str, float]) == (1, "a", 2.0)

    # Lists of lists, dictionaries of dictionaries, etc.
    assert validator.instantiate([[1, 2], [3, 4]], List[List[int]]) == [[1, 2], [3, 4]]
    assert validator.instantiate({"a": {"b": 2}, "c": {"d": 4}}, Dict[str, Dict[str, int]]) == {"a": {"b": 2},
                                                                                                "c": {"d": 4}}
    # Lists that can contain multiple types of items.
    assert validator.instantiate([1, "a", 2.0], List[Union[int, str, float]]) == [1, "a", 2.0]
    assert validator.instantiate({"a": 1, "b": "c"}, Dict[str, Union[int, str]]) == {"a": 1, "b": "c"}

    # Items that can be None.
    assert validator.instantiate([1, None, 2], List[Optional[int]]) == [1, None, 2]

    # Combinations of the above.
    data = {"a": [1, 2, 3], "b": [4, "c", 6]}
    assert validator.instantiate(data, Dict[str, List[Union[int, str]]]) == data

    # Data that doesn't match the provided type should raise an error.
    try:
        validator.instantiate([1, 2, "a"], List[int])
        assert False, "Expected a TypeError"
    except TypeError:
        pass

    # Ensure that custom classes without generics work as expected.
    class SimpleClass:
        def __init__(self, x):
            self.x = x

    assert isinstance(validator.instantiate({"x": 10}, SimpleClass), SimpleClass)

    # ABCs like Sequence, Iterable, Mapping instead of concrete types like list, tuple, dict.
    assert validator.instantiate([1, 2, 3], Sequence[int]) == [1, 2, 3]
    assert validator.instantiate((1, 2, 3), Iterable[int]) == (1, 2, 3)
    assert validator.instantiate({"a": 1, "b": 2}, Mapping[str, int]) == {"a": 1, "b": 2}

    # Deque
    result = validator.instantiate([1, 2, 3], deque)
    assert isinstance(result, deque) and list(result) == [1, 2, 3]

    # FrozenSet
    result = validator.instantiate([1, 2, 3], FrozenSet[int])
    assert isinstance(result, frozenset) and list(result) == [1, 2, 3]

    # DefaultDict
    result = validator.instantiate({"a": 1, "b": 2}, DefaultDict[str, int])
    assert isinstance(result, defaultdict) and dict(result) == {"a": 1, "b": 2}

    # Counter
    #result = validator.instantiate({"a": 1, "b": 2}, Counter[str])
    #assert isinstance(result, Counter) and dict(result) == {"a": 1, "b": 2}

    # OrderedDict
    result = validator.instantiate({"a": 1, "b": 2}, OrderedDict[str, int])
    assert isinstance(result, OrderedDict) and dict(result) == {"a": 1, "b": 2}

@pytest.fixture
def validator():
    return Validator()

def test_extended_generics(validator):

    test_extended_list(validator)
    test_extended_tuple(validator)
    test_extended_dict(validator)
    test_extended_set(validator)


def test_extended_set(validator):
    class ExtendedSet(Set[int]):
        pass

    # We are using a list here because that is how they are represented in JSON.
    extended_set = validator.instantiate([1, 2, 3], ExtendedSet)
    assert isinstance(extended_set, ExtendedSet)
    # Test type coercion
    extended_set = validator.instantiate([1, "2", 3], ExtendedSet)
    assert isinstance(extended_set, ExtendedSet)
    # Lets now try and violate the inherited type
    try:
        validator.instantiate([1, "a", 3], ExtendedSet)
        assert False, "Expected a ValueError"
    except TypeError:
        pass


def test_extended_dict(validator):
    class ExtendedDict(Dict[str, int]):
        pass

    extended_dict = validator.instantiate({"a": 1, "b": 2}, ExtendedDict)
    assert isinstance(extended_dict, ExtendedDict)
    # Test type coercion
    extended_dict = validator.instantiate({"a": "1", "b": 2}, ExtendedDict)
    assert isinstance(extended_dict, ExtendedDict)
    # Lets now try and violate the inherited type
    try:
        validator.instantiate({4: 1, "b": 2}, ExtendedDict)
        assert False, "Expected a ValueError"
    except TypeError:
        pass


def test_extended_tuple(validator):
    class ExtendedTuple(Tuple[int, str]):
        pass

    # We are using a list here because that is how they are represented in JSON.
    extended_tuple = validator.instantiate([1, "a"], ExtendedTuple)
    assert isinstance(extended_tuple, ExtendedTuple)
    # Test type coercion
    extended_tuple = validator.instantiate(["1", 1], ExtendedTuple)
    assert isinstance(extended_tuple, ExtendedTuple)
    # Lets now try and violate the inherited type
    try:
        validator.instantiate(['a', 2], ExtendedTuple)
        assert False, "Expected a ValueError"
    except TypeError:
        pass


def test_extended_list(validator):
    class ExtendedList(List[int]):
        pass

    extended_list = validator.instantiate([1, 2, 3], ExtendedList)
    assert isinstance(extended_list, ExtendedList)
    # Lets now try and violate the inherited type
    try:
        validator.instantiate([1, "a"], ExtendedList)
        assert False, "Expected a ValueError"
    except TypeError:
        pass


if __name__ == "__main__":
    test_instantiate_pydantic()
    test_instantiate_dataclass()
    test_primitives()
    test_generics()
    test_extended_generics()


from typing import Dict

from tanuki.language_models.embedding_api_abc import Embedding_API
from tanuki.language_models.openai_api import OpenAI_API
from tanuki.models.embedding import Embedding
from tanuki.models.embedding_model_output import EmbeddingModelOutput
from tanuki.models.function_description import FunctionDescription


class EmbeddingModelManager(object):
    def __init__(self, function_modeler):
        self.function_modeler = function_modeler
        self.api_models: Dict[str, Embedding_API] = {"openai": OpenAI_API()}

    def get_embedding_case(self, args, function_description: FunctionDescription, kwargs, examples=None):
        # example_input = f"Examples:{examples}\n" if examples else ""
        content = f"Name: {function_description.name}\nArgs: {args}\nKwargs: {kwargs}"
        return content

    def __call__(self,
                 args,
                 function_description,
                 kwargs) -> Embedding:
        prompt = self.get_embedding_case(args, function_description, kwargs)
        embedding_response = self.api_models["openai"].embed([prompt])[0]
        embedding: Embedding = function_description.output_type_hint(embedding_response.embedding)


        # We don't need to postprocess the embedding for now, because saving them offers no information that we can use
        # to improve the model.

        #is_distilled_model = False  # TODO: check if distilled model
        #output = EmbeddingModelOutput(embedding, is_distilled_model)
        #if not output.distilled_model:
        #    self.function_modeler.postprocess_embeddable_datapoint(function_description.__hash__(),
        #                                                           function_description,
        #                                                           embedding)

        return embedding



import time
from tempfile import TemporaryDirectory

import pytest

from tanuki.models.function_example import FunctionExample
from tanuki.trackers.filesystem_buffered_logger import FilesystemBufferedLogger


@pytest.fixture(params=[FilesystemBufferedLogger])
def logger(request):
    with TemporaryDirectory() as temp_dir:
        global ALIGN_FILE_NAME
        ALIGN_FILE_NAME = temp_dir
        yield request.param("test")


#def test_load_log_align(logger):
#    runs = 100000
#
#    start_time = time.time()
#    for i in range(runs):
#        example = FunctionExample((i,), {}, i * 2)
#        logger.log_align(str(i), example)
#    elapsed_time = time.time() - start_time
#
#    print(f"Time taken for {logger.__class__.__name__}: {elapsed_time} seconds")

#def test_patch_many_functions(logger):
#    runs = 10000
#
#    start_time = time.time()
#    for i in range(runs):
#        example = FunctionExample((i,), {}, i * 2)
#        logger.log_patch(str(i), example)
#    elapsed_time = time.time() - start_time
#
#    print(f"Time taken for {logger.__class__.__name__} to patch {runs} functions: {elapsed_time} seconds")
#
def test_patch_one_function_many_times():
    runs = 100
    logger = FilesystemBufferedLogger("test")
    logger.bloom_filter = logger.create_bloom_filter()
    start_time = time.time()
    for i in range(runs):
        example = FunctionExample((i,), {}, i * 2)
        before_bit_array = logger.bloom_filter.bit_array.copy()

        logger.log_symbolic_patch("test", example)

        after_bit_array = logger.bloom_filter.bit_array
        is_same = before_bit_array == after_bit_array

        assert is_same == False

    logger.save_bloom_filter()
    elapsed_time = time.time() - start_time

    print(f"Time taken for {logger.__class__.__name__} to patch function {runs} times: {elapsed_time} seconds")
    print(f"Hits: {logger.hit_count}, Misses: {logger.miss_count}")

    start_time = time.time()
    for i in range(runs):
        example = FunctionExample((i,), {}, i * 2)

        logger.log_symbolic_patch("test", example)

        after_bit_array = logger.bloom_filter.bit_array
        #is_same = before_bit_array == after_bit_array


    logger.save_bloom_filter()
    elapsed_time = time.time() - start_time

    print(f"Time taken for {logger.__class__.__name__} to patch function {runs} times: {elapsed_time} seconds")
    print(f"Hits: {logger.hit_count}, Misses: {logger.miss_count}")

if __name__ == "__main__":
    #test_patch_many_functions(BufferedLogger())
    test_patch_one_function_many_times()


import ast
import inspect
import json
import logging
import os
import sys
import textwrap
from functools import wraps
from typing import Optional, Union, Any
from unittest.mock import patch as mock_patch

import requests

from tanuki.assertion_visitor import AssertionVisitor
from tanuki.function_modeler import FunctionModeler
from tanuki.language_models.embedding_model_manager import EmbeddingModelManager
from tanuki.language_models.language_model_manager import LanguageModelManager
from tanuki.models.embedding import Embedding
from tanuki.models.function_description import FunctionDescription
from tanuki.models.function_example import FunctionExample
from tanuki.models.function_type import FunctionType
from tanuki.register import Register
from tanuki.trackers.filesystem_buffered_logger import FilesystemBufferedLogger
from tanuki.utils import get_key
from tanuki.validator import Validator


# Define a new level
def _log_align(self, func_hash, *args, **kws):
    if self.isEnabledFor(ALIGN_LEVEL_NUM):
        args, kwargs, output = args
        kwargs['align'] = True
        example = FunctionExample(args, kwargs, output)

        # Define a safe directory within the project for logs
        # (You can make this configurable if needed)
        log_directory = os.path.join(os.getcwd(), ALIGN_FILE_NAME)

        # Ensure the directory exists
        if not os.path.exists(log_directory):
            try:
                os.makedirs(log_directory)
            except OSError as e:
                self.error(f"Failed to create log directory: {e}")
                return

        # Write to the file
        log_file_path = os.path.join(log_directory, func_hash)
        try:
            with open(log_file_path, "a") as f:
                f.write(str(example.__dict__) + "\n")
        except IOError as e:
            self.error(f"Failed to write to log file: {e}")


# Set up logging with custom logger
def logger_factory(name):
    return FilesystemBufferedLogger(name)


ALIGN_LEVEL_NUM = 15
PATCH_LEVEL_NUM = 14
ALIGN_FILE_NAME = ".align"

alignable_functions = {}

# Set up basic configuration
logging.setLoggerClass(FilesystemBufferedLogger)
logging.addLevelName(ALIGN_LEVEL_NUM, "ALIGN")
logging.addLevelName(PATCH_LEVEL_NUM, "PATCH")
logging.basicConfig(level=ALIGN_LEVEL_NUM)
logger = logger_factory(__name__)
# currently only use buffered logger as default
function_modeler = FunctionModeler(data_worker=logger)
language_modeler = LanguageModelManager(function_modeler)
embedding_modeler = EmbeddingModelManager(function_modeler)
telemetry_enabled: bool = True


@staticmethod
def _load_alignments(func_hash: str):
    function_modeler.load_symbolic_align_statements(func_hash)

@staticmethod
def _anonymous_usage(*args, **kwargs):
    """
    Post anonymously to the usage server so we know what configs are commonly used in the project.
    :return:
    """
    if not telemetry_enabled:
        return
    try:
        requests.post('https://idhhnusnhkkjkpwkm1fr.monkeypatch.ai/telemetry', data=json.dumps(kwargs))
    except:
        pass

@staticmethod
def align(test_func):
    """
    Decorator to align a function.

    By adding the @align decorator to a function, we can declare the desired input-output
    behaviour of the patched functions using assertions.

    :param test_func:
    :return:
    """

    @wraps(test_func)
    def wrapper(*args, **kwargs):
        source = textwrap.dedent(inspect.getsource(test_func))
        tree = ast.parse(source)
        _locals = locals()

        # We are handling symbolic and embeddable functions differently, as they have different semantics during
        # the alignment process.

        patch_symbolic_funcs = Register.functions_to_patch(type=FunctionType.SYMBOLIC)
        patch_embeddable_funcs = Register.functions_to_patch(type=FunctionType.EMBEDDABLE)
        visitor = AssertionVisitor(_locals,
                                   patch_symbolic_funcs=patch_symbolic_funcs,
                                   patch_embeddable_funcs=patch_embeddable_funcs)
        visitor.visit(tree)

        # Get the mocked behaviours from analyzing the AST of the aligned function
        mock_behaviors = visitor.mocks

        # Negative examples (i.e. embeddable function examples that should have maximum distance in the embedding space)
        mock_negatives = visitor.negative_mocks

        if args:
            instance = args[0]
            args = args[1:]
        else:
            instance = None

        def extract_attributes(result):
            attributes = {}

            # If the result is a list, get its length
            if isinstance(result, list):
                attributes['length'] = len(result)

            # If the result is a dictionary, get its keys (or any other attributes)
            elif isinstance(result, dict):
                attributes['keys'] = list(result.keys())

            return attributes

        def create_mock_func(instance: Optional,
                             func_name: str,
                             description: FunctionDescription):

            def mock_func(*args, **kwargs):
                hashed_description = description.__hash__()

                function_type, func = Register.get(func_name)

                # If we are aligning a function that returns an embedding,
                # we need to ensure both sides of the equality are future embeddings,
                # as it is nonsensical to declare that an embedding should 'be' an object or a string, etc.
                if function_type == FunctionType.EMBEDDABLE:
                    key = get_key(args, kwargs)
                    mocked_embedding = mock_behaviors.get(key, None)

                    # Find positive examples by matching the mocked embedding with identical embeddings in the values
                    # of the mock_behaviors dictionary
                    mock_positives_list = []
                    for k, v in mock_behaviors.items():
                        if v == mocked_embedding and k != key:
                            mock_positives_list.append(k)
                    equivalent_mocks = mock_positives_list
                    negative_mocks = list(mock_negatives.values())
                    function_modeler.save_embeddable_align_statements(hashed_description,
                                                                             args,
                                                                             kwargs,
                                                                             equivalent_mocks,
                                                                             negative_mocks)
                    return mocked_embedding
                else:
                    # If we are aligning a function that returns an object
                    if not instance:
                        result = func(*args, **kwargs)
                    else:
                        result = func(instance, *args, **kwargs)

                    # Extract attributes from the result
                    attributes = extract_attributes(result)
                    for attr_name, attr_value in attributes.items():
                        # If the attribute is a list, get its length
                        if isinstance(attr_value, list):
                            attributes[attr_name] = len(attr_value)

                    key = get_key(args, kwargs)
                    mocked_behaviour = mock_behaviors.get(key, None)
                    function_modeler.save_symbolic_align_statements(hashed_description, args, kwargs,
                                                                           mocked_behaviour)
                    return mocked_behaviour

            return mock_func


        # Identify all functions that need to be patched based on mock_behaviors
        if instance:
            function_names_to_patch = Register.function_names_to_patch(instance)#, type=FunctionType.SYMBOLIC)
            functions_descriptions = [Register.load_function_description_from_name(instance, func_name)
                                      for func_name in function_names_to_patch]
        else:
            function_names_to_patch = Register.function_names_to_patch()#type=FunctionType.SYMBOLIC)
            functions_descriptions = [Register.load_function_description_from_name(func_name)
                                      for func_name in function_names_to_patch]

        patched_func = test_func
        for desc, func in zip(functions_descriptions, function_names_to_patch):
            mock_function = create_mock_func(instance, func, desc)
            module_name = sys.modules[test_func.__module__].__name__

            if instance:
                patched_func = mock_patch.object(instance, func, new=mock_function)(patched_func)
            else:
                patched_func = mock_patch(f'{module_name}.{func}', new=mock_function)(patched_func)

        # Get the signature of the function
        sig = inspect.signature(test_func)

        if sig.parameters:
            first_param_name = next(iter(sig.parameters))

            # If the instance is the "self" or the name of the first parameter,
            # then pass it as the first argument
            if first_param_name in ['self', 'cls'] or first_param_name == instance:
                return patched_func(instance, *args, **kwargs)
            else:
                return patched_func(*args, **kwargs)
        else:
            return patched_func(*args, **kwargs)

    return wrapper

@staticmethod
def generate_from_embedding_model_manager(function_description):
    choice_parsed = []
    instantiated = function_description.output_type_hint(choice_parsed)
    return instantiated

@staticmethod
def patch(patchable_func=None,
          environment_id: int = 0,
          ignore_finetune_fetching: bool = False,
          ignore_finetuning: bool = False,
          ignore_data_storage: bool = False
          ):
    """
    The main decorator for patching a function.
    args:
        patchable_func: The function to be patched, should be always set to none. This is used here to allow for keyword arguments or no arguments to be passed to the decorator
        environment_id (int): The environment id. Used for fetching correct finetuned models
        ignore_finetune_fetching (bool): Whether to ignore fetching finetuned models.
            If set to True, during the first call openai will not be queried for finetuned models, which reduces initial startup latency
        ignore_finetuning (bool): Whether to ignore finetuning the models altogether. If set to True the teacher model will always be used.
            The data is still saved however if in future would need to use finetuning
        ignore_data_storage (bool): Whether to ignore storing the data.
            If set to True, the data will not be stored in the finetune dataset and the align statements will not be saved
            This improves latency as communications with data storage is minimised
    """

    def wrap(test_func):
        @wraps(test_func)
        def wrapper(*args, **kwargs) -> Union[Embedding, Any]:
            validator = Validator()
            function_description: FunctionDescription = Register.load_function_description(test_func)

            # If the function is expected to return an embedding, we choose the embedding API, rather than an LLM.
            if inspect.isclass(function_description.output_type_hint) and \
                    issubclass(function_description.output_type_hint, Embedding):
                instantiated: Embedding = embedding_modeler(args, function_description, kwargs)
            else:
                # If the function is expected to return a choice, we choose the LLM API.
                instantiated: Any = language_modeler(args, function_description, kwargs, validator)

            return instantiated  # test_func(*args, **kwargs)

        _anonymous_usage(logger=logger.name)
        function_description = Register.load_function_description(test_func)
        func_hash = function_description.__hash__()
        function_modeler.environment_id = environment_id
        if ignore_finetuning:
            function_modeler.execute_finetune_blacklist.append(func_hash)
        if ignore_finetune_fetching:
            function_modeler.check_finetune_blacklist.append(func_hash)
        if ignore_data_storage:
            function_modeler.store_data_blacklist.append(func_hash)
        _load_alignments(func_hash)

        wrapper._is_alignable = True
        Register.add_function(test_func, function_description)
        return wrapper

    if callable(patchable_func):
        func = patchable_func
        return wrap(func)
    if patchable_func is not None:
        raise TypeError(
            "The first argument to patch must not be specified. Please use keyword arguments or specify the first argument as None")
    return wrap



from dataclasses import dataclass
from typing import TypeVar

from tanuki.models.embedding import Embedding

T = TypeVar('T')

@dataclass()
class EmbeddingModelOutput:
    generated_response: Embedding[T]
    #suitable_for_finetuning: bool
    distilled_model: bool



import openai
import os
from dotenv import load_dotenv
from pydantic import BaseModel
from typing import Optional

load_dotenv()

import tanuki
from utils import scrape_url


openai.api_key = os.getenv("OPENAI_API_KEY")


class Job(BaseModel):
    position: str
    company: str
    location: str


@tanuki.patch
def extract_job(content: str) -> Optional[Job]:
    """
    Examine the content string and extract the job details for the position title, company, and location.
    """


@tanuki.align
def align_extract_job() -> None:
    print("Aligning...")
    job = "\n\n\n\n\n\n\n\n\nShip broker\nFuentes, Walls and Castro\n\n\n\n\n        Michelleville, AP\n      \n\n2021-04-08\n\n\n\nLearn\nApply\n\n\n"
    assert extract_job(job) == Job(
        position="Ship broker",
        company="Fuentes, Walls and Castro",
        location="Michelleville, AP",
    )


if __name__ == '__main__':

    # Align the function
    align_extract_job()

    # Web scrape the url and extract the list of jobs
    url = "https://realpython.github.io/fake-jobs/"
    contents = scrape_url(url=url, class_name="card")

    # Process the job blocks using Tanuki (only sampling a couple for demo purposes)
    jobs = []
    for content in contents[1:3]:
        jobs.append(extract_job(content))
    print(jobs)



import ast
import datetime
import io
import json
from typing import List, Tuple, Dict

import openai

from tanuki.constants import EXAMPLE_ELEMENT_LIMIT, PATCHES, SYMBOLIC_ALIGNMENTS, POSITIVE_EMBEDDABLE_ALIGNMENTS, \
    NEGATIVE_EMBEDDABLE_ALIGNMENTS
from tanuki.models.function_description import FunctionDescription
from tanuki.models.function_example import FunctionExample
from tanuki.trackers.dataset_worker import DatasetWorker
from tanuki.utils import approximate_token_count, prepare_object_for_saving, encode_int, decode_int
import copy


class FunctionModeler(object):
    """
    This class manages the registered function models and their datasets
    comprised of symbolic and embeddable alignments, and symbolic and embeddable patches
    """

    def __init__(self, data_worker: DatasetWorker, environment_id=0) -> None:
        self.function_configs = {}
        self.data_worker = data_worker
        self.distillation_token_limit = 3000  # the token limit for finetuning
        self.symbolic_align_buffer = {}
        self.embeddable_align_buffer = {}
        self._get_datasets()
        self.environment_id = environment_id
        self.check_finetune_blacklist = []
        self.execute_finetune_blacklist = []
        self.store_data_blacklist = []

    def _get_dataset_info(self, dataset_type, func_hash, type="length"):
        """
        Get the dataset size for a function hash
        """
        return self.data_worker.load_dataset(dataset_type, func_hash, return_type=type)

    def _get_datasets(self):
        """
        Get the existing datasets from the data worker
        """
        self.dataset_sizes = self.data_worker.load_existing_datasets()

    def save_embeddable_align_statements(self,
                                         function_hash: str,
                                         args,
                                         kwargs,
                                         positive_pairs: List[Tuple[List, Dict]],
                                         negative_pairs: List[Tuple[List, Dict]]):
        """
        Save the contrastive align statements for the embeddable function.
        Do not save if the function hash is in the store data blacklist

        Args:
            function_hash: A unique hash for the function
            args: The arguments of the function
            kwargs: The keyword arguments of the function
            positive_pairs: A list of the other function invocations that are should have equivalent embeddings
            negative_pairs: A list of the other function invocations that are should have different embeddings
        """
        # prepare args and kwargs for saving
        copy_args = copy.deepcopy(args)
        copy_kwargs = copy.deepcopy(kwargs)
        parsed_args = prepare_object_for_saving(copy_args)
        parsed_kwargs = prepare_object_for_saving(copy_kwargs)

        # prepare positive pairs for saving
        parsed_positive_pairs = []
        for pair in positive_pairs:
            copy_pair = copy.deepcopy(pair)
            parsed_pair = prepare_object_for_saving(copy_pair)
            parsed_positive_pairs.append(parsed_pair)

        # prepare negative pairs for saving
        parsed_negative_pairs = []
        for pair in negative_pairs:
            copy_pair = copy.deepcopy(pair)
            parsed_pair = prepare_object_for_saving(copy_pair)
            parsed_negative_pairs.append(parsed_pair)

        # save the contrastive pairs
        for pair in parsed_positive_pairs:
            self._save_contrastive_alignment_pair(function_hash, parsed_args, parsed_kwargs, pair, positive=True)
        for pair in parsed_negative_pairs:
            self._save_contrastive_alignment_pair(function_hash, parsed_args, parsed_kwargs, pair, positive=False)

    def _save_contrastive_alignment_pair(self, function_hash: str, args, kwargs, pair, positive=True):
        """
        Save a contrastive pair
        """
        example = FunctionExample(args, kwargs, pair)
        if function_hash not in self.store_data_blacklist:
            successfully_saved, new_datapoint = self.data_worker.log_embeddable_align(function_hash, example, positive)
        else:
            successfully_saved = False
            new_datapoint = True
        if successfully_saved:
            if positive:
                if function_hash in self.dataset_sizes[POSITIVE_EMBEDDABLE_ALIGNMENTS]:
                    self.dataset_sizes[POSITIVE_EMBEDDABLE_ALIGNMENTS][function_hash] += 1
                else:
                    self.dataset_sizes[POSITIVE_EMBEDDABLE_ALIGNMENTS][function_hash] = 1
            if not positive:
                if function_hash in self.dataset_sizes[NEGATIVE_EMBEDDABLE_ALIGNMENTS]:
                    self.dataset_sizes[NEGATIVE_EMBEDDABLE_ALIGNMENTS][function_hash] += 1
                else:
                    self.dataset_sizes[NEGATIVE_EMBEDDABLE_ALIGNMENTS][function_hash] = 1

        if new_datapoint:
            # update align buffer
            if function_hash not in self.embeddable_align_buffer:
                self.embeddable_align_buffer[function_hash] = bytearray()
            self.embeddable_align_buffer[function_hash].extend(str(example.__dict__).encode('utf-8') + b'\r\n')

    def save_symbolic_align_statements(self, function_hash, args, kwargs, output):
        """
        Save the align statements and add to the align buffer
        Do not save if the function hash is in the store data blacklist
        Then just add the datapoints to the align buffer
        """
        # prepare output for saving and later parsing
        # make a deepcopy of the output to avoid changing the original object
        copy_output = copy.deepcopy(output)
        parsed_output = prepare_object_for_saving(copy_output)

        # prepare args and kwargs for saving
        copy_args = copy.deepcopy(args)
        copy_kwargs = copy.deepcopy(kwargs)
        parsed_args = prepare_object_for_saving(copy_args)
        parsed_kwargs = prepare_object_for_saving(copy_kwargs)

        example = FunctionExample(parsed_args, parsed_kwargs, parsed_output)
        if function_hash not in self.store_data_blacklist:
            successfully_saved, new_datapoint = self.data_worker.log_symbolic_align(function_hash, example)
        else:
            successfully_saved = False
            new_datapoint = True
        if successfully_saved:
            if function_hash in self.dataset_sizes[SYMBOLIC_ALIGNMENTS]:
                self.dataset_sizes[SYMBOLIC_ALIGNMENTS][function_hash] += 1
            else:
                self.dataset_sizes[SYMBOLIC_ALIGNMENTS][function_hash] = 1

        if new_datapoint:
            # update align buffer
            if function_hash not in self.symbolic_align_buffer:
                self.symbolic_align_buffer[function_hash] = bytearray()
            self.symbolic_align_buffer[function_hash].extend(str(example.__dict__).encode('utf-8') + b'\r\n')

    def save_symbolic_datapoint(self, func_hash, example):
        """
        Save datapoint to the training data
        """
        written_datapoints = self.data_worker.log_symbolic_patch(func_hash, example)
        for func_hash, datapoints in written_datapoints.items():
            if func_hash in self.dataset_sizes[PATCHES]:
                # if the dataset size is -1, it means we havent read in the dataset size yet
                if self.dataset_sizes[PATCHES][func_hash] == -1:
                    self.dataset_sizes[PATCHES][func_hash] = self._get_dataset_info(PATCHES, func_hash, type="length")
                else:
                    self.dataset_sizes[PATCHES][func_hash] += datapoints
            else:
                self.dataset_sizes[PATCHES][func_hash] = datapoints
        return len(written_datapoints) > 0

    def get_symbolic_alignments(self, func_hash, max=20):
        """
        Get all symbolic aligns for a function hash
        """

        if func_hash not in self.symbolic_align_buffer:
            return []

        buffer = self.symbolic_align_buffer[func_hash]
        return self._get_examples_from_alignment_buffer(buffer, max)

    def get_embeddable_alignments(self, func_hash, max=20):
        """
        Get all embeddable aligns for a function hash
        """

        if func_hash not in self.embeddable_align_buffer:
            return []

        buffer = self.embeddable_align_buffer[func_hash]
        return self._get_examples_from_alignment_buffer(buffer, max)

    def _get_examples_from_alignment_buffer(self, buffer, max=20):
        """
        Get examples from a buffer
        """

        split_buffer = bytes(buffer).split(b"\n")

        # byte array of stringed python dicts into dict objects
        example_set = set()
        for example in split_buffer:
            if example == b"":
                continue
            example_set.add(example)

        # easy and straightforward way to get nr of words (not perfect but doesnt need to be)
        # Can do the proper way of tokenizing later, it might be slower and we dont need 100% accuracy
        example_element_limit = EXAMPLE_ELEMENT_LIMIT

        examples = []
        for example_bytes in split_buffer:
            if example_bytes in example_set:
                nr_of_elements = approximate_token_count(example_bytes)
                example_element_limit -= nr_of_elements
                if example_element_limit < 0:
                    break
                example = example_bytes.decode('utf-8')
                # json load the example
                try:
                    example = json.loads(example)
                except:
                    example = ast.literal_eval(example)
                examples.append(example)
                example_set.remove(example_bytes)

        return list(examples)[:max]

    def load_symbolic_align_statements(self, function_hash):
        """
        Load all align statements
        First check the data storage blacklist,
        if the func hash is in the blacklist, then set the dataset size to 0 and the align buffer to empty bytearray
        """
        if function_hash in self.store_data_blacklist:
            self.dataset_sizes[SYMBOLIC_ALIGNMENTS][function_hash] = 0
            self.symbolic_align_buffer[function_hash] = bytearray()

        elif function_hash not in self.symbolic_align_buffer:
            dataset_size, align_dataset = self._get_dataset_info(SYMBOLIC_ALIGNMENTS, function_hash, type="both")
            if align_dataset:
                self.symbolic_align_buffer[function_hash] = bytearray(align_dataset)
            self.dataset_sizes[SYMBOLIC_ALIGNMENTS][function_hash] = dataset_size

    def postprocess_symbolic_datapoint(self, func_hash, function_description, example, repaired=True):
        """
        Postprocess the datapoint
        First check if the datapoint should be added to the training data
        Add the datapoint if it should be added
        Then check if the function should be finetuned and execute finetuning if it should
        """
        try:
            if func_hash not in self.store_data_blacklist:
                added = self.save_symbolic_datapoint(func_hash, example)
                if added:
                    self._update_datapoint_config(repaired, func_hash)
        except Exception as e:
            print(e)
            print("Could not add datapoint to training data")
        if func_hash not in self.execute_finetune_blacklist:
            self.check_for_finetuning(function_description, func_hash)

    def load_function_config(self, func_hash, function_description):
        """
        Load the config file for a function hash
        """
        config, default = self.data_worker.load_function_config(func_hash)
        if default and func_hash not in self.check_finetune_blacklist:
            finetuned, finetune_config = self._check_for_finetunes(function_description)
            if finetuned:
                config = finetune_config
        self.function_configs[func_hash] = config
        return config

    def _check_for_finetunes(self, function_description: FunctionDescription) -> Tuple[bool, Dict]:
        # This here should be discussed, what's the bestd way to do it
        # hash the function_hash into 16 characters (to embed it into the name of OpenAI finetunes, for later retrieval)

        finetune_hash = function_description.__hash__(purpose="finetune") + encode_int(self.environment_id)
        # List 10 fine-tuning jobs
        finetunes = openai.FineTuningJob.list(limit=1000)
        # Check if the function_hash is in the fine-tuning jobs
        # the finetunes are in chronological order starting from newest
        # So this gets the latest finetune
        for finetune in finetunes["data"]:
            # check if the finetune hash is in the fine-tuned model name
            if finetune["status"] == "succeeded" and finetune_hash in finetune["fine_tuned_model"]:
                try:
                    config = self._construct_config_from_finetune(finetune_hash, finetune)
                    # save the config
                    self.data_worker.update_function_config(function_description.__hash__(), config)
                    return True, config
                except:
                    return False, {}

        return False, {}

    def _construct_config_from_finetune(self, finetune_hash, finetune):
        model = finetune["fine_tuned_model"]
        # get the ending location of finetune hash in the model name
        finetune_hash_end = model.find(finetune_hash) + len(finetune_hash)
        # get the next character after the finetune hash
        next_char = model[finetune_hash_end]
        # get the number of training runs
        nr_of_training_runs = decode_int(next_char) + 1
        nr_of_training_points = (2 ** (nr_of_training_runs - 1)) * 200
        config = {
            "distilled_model": model,
            "current_model_stats": {
                "trained_on_datapoints": nr_of_training_points,
                "running_faults": []},
            "last_training_run": {"trained_on_datapoints": nr_of_training_points},
            "current_training_run": {},
            "teacher_models": ["gpt-4", "gpt-4-32k"],  # currently supported teacher models
            "nr_of_training_runs": nr_of_training_runs}

        return config

    def get_models(self, function_description):
        """
        Return the current model from the config file
        """
        func_hash = function_description.__hash__()
        if func_hash in self.function_configs:
            func_config = self.function_configs[func_hash]
        else:
            func_config = self.load_function_config(func_hash, function_description)

        # for backwards compatibility
        if "distilled_model" not in func_config:
            if func_config["current_model"] in func_config["teacher_models"]:
                distilled_model = ""
            else:
                distilled_model = func_config["current_model"]
        else:
            distilled_model = func_config["distilled_model"]

        return distilled_model, func_config["teacher_models"]

    def _update_datapoint_config(self, repaired, func_hash):
        """
        Update the config to reflect the new datapoint in the training data
        First adds 1 to the current datapoints
        Then updates running faults depending if priority is True or not and takes last 100
        Then checks the revert condition, i.e if last 10 datapoints are 50% faulty
        Finally updates the config file 
        Args:
           priority (bool): whether the datapoint was fixed by the teacher model/should be added to the training data
        """
        try:
            if repaired:
                self.function_configs[func_hash]["current_model_stats"]["running_faults"].append(1)
            else:
                self.function_configs[func_hash]["current_model_stats"]["running_faults"].append(0)
            # take the last 100 datapoints
            self.function_configs[func_hash]["current_model_stats"]["running_faults"] = \
                self.function_configs[func_hash]["current_model_stats"]["running_faults"][-100:]

            # check if the last 10 datapoints are 50% faulty, this is the switch condition
            if sum(self.function_configs[func_hash]["current_model_stats"]["running_faults"][-10:]) / 10 > 0.5:
                self.function_configs[func_hash]["distilled_model"] = ""
                self.function_configs[func_hash]["current_model_stats"]["trained_on_datapoints"] = 0
                self.function_configs[func_hash]["current_model_stats"]["running_faults"] = []
            self._update_config_file(func_hash)

        except Exception as e:
            print(e)
            print("Could not update config file")
            pass

    def _update_config_file(self, func_hash):
        self.data_worker.update_function_config(func_hash, self.function_configs[func_hash])

    def check_for_finetuning(self, function_description, func_hash):
        """
        Check for finetuning status
        If already finetuning, check for finetuning status
        If not finetuning, check for finetuning condition and execute finetuning if condition is met
        """
        try:
            # check if already finetuning
            if "job_id" in self.function_configs[func_hash]["current_training_run"]:
                # check for job status
                self._check_finetuning_status(func_hash)
            else:
                # check for finetuning condition
                if self._check_finetuning_condition(func_hash):
                    self._execute_finetuning(function_description, func_hash)
        except Exception as e:
            print(e)
            print("Error checking for finetuning")

    def _check_finetuning_condition(self, func_hash):
        """
        Check if the finetuning condition is met
        Currently finetuning condition is dependent on the number of symbolic datapoints since last finetuning
        """
        if func_hash not in self.function_configs:
            return False

        training_threshold = (2 ** self.function_configs[func_hash]["nr_of_training_runs"]) * 200

        align_dataset_size = self.dataset_sizes[SYMBOLIC_ALIGNMENTS][func_hash] if func_hash in self.dataset_sizes[
            SYMBOLIC_ALIGNMENTS] else 0
        patch_dataset_size = self.dataset_sizes[PATCHES][func_hash] if func_hash in self.dataset_sizes[PATCHES] else 0

        if patch_dataset_size == -1:
            # if havent read in the patch dataset size, read it in
            patch_dataset_size = self._get_dataset_info(PATCHES, func_hash, type="length")
            self.dataset_sizes[PATCHES][func_hash] = patch_dataset_size

        return (patch_dataset_size + align_dataset_size) > training_threshold

    def _execute_finetuning(self, function_description, func_hash):
        """
        Execute the finetuning
        First create the OpenAI compatible dataset with jsonL file and upload it
        Then submit the OpenAI finetuning job
        Finally update the config file to reflect the new finetuning job as current
        """
        # get function description
        function_string = str(function_description.__dict__.__repr__() + "\n")

        # get the align dataset
        align_dataset = self._get_dataset_info(SYMBOLIC_ALIGNMENTS, func_hash, type="dataset")
        if not align_dataset:
            align_dataset = ""
        else:
            align_dataset = align_dataset.decode('utf-8')

        # get the patch dataset
        patch_dataset = self._get_dataset_info(PATCHES, func_hash, type="dataset")
        if not patch_dataset:
            patch_dataset = ""
        else:
            patch_dataset = patch_dataset.decode('utf-8')

        if align_dataset == "" and patch_dataset == "":
            return

        dataset = align_dataset + patch_dataset

        dataset.replace("\\n", "[SEP_TOKEN]")
        dataset = dataset.split("\n")
        dataset = [x.replace("[SEP_TOKEN]", "\\n") for x in dataset if x != ""]
        # read in the dataset file
        dataset = [ast.literal_eval(x) for x in dataset]
        #
        # create the openai dataset
        instruction = "You are given below a function description and input data. The function description of what the function must carry out can be found in the Function section, with input and output type hints. The input data can be found in Input section. Using the function description, apply the function to the Input and return a valid output type, that is acceptable by the output_class_definition and output_class_hint. Return None if you can't apply the function to the input or if the output is optional and the correct output is None.\nINCREDIBLY IMPORTANT: Only output a JSON-compatible string in the correct response format."
        finetuning_dataset = [{"messages": [
            {
                "role": "system",
                "content": f"You are a skillful and accurate language model, who applies a described function on input data. Make sure the function is applied accurately and correctly and the outputs follow the output type hints and are valid outputs given the output types."
            },
            {"role": "user",
             "content": f"{instruction}\nFunction: {function_string}---\nInputs:\nArgs: {x['args']}\nKwargs: {x['kwargs']}\nOutput:"},
            {"role": "assistant", "content": str(x['output']) if x['output'] is not None else "None"}]}
            for x in dataset]

        # Create an in-memory text stream
        temp_file = io.StringIO()
        # Write data to the stream
        for idx, item in enumerate(finetuning_dataset):
            temp_file.write(json.dumps(item))
            if idx != len(finetuning_dataset) - 1:
                temp_file.write("\n")

        # Reset the stream position to the beginning
        temp_file.seek(0)

        # create the finetune hash
        finetune_hash = function_description.__hash__(purpose="finetune")
        nr_of_training_runs = self.function_configs[func_hash]["nr_of_training_runs"]
        finetune_hash += encode_int(self.environment_id)
        finetune_hash += encode_int(nr_of_training_runs)

        # Use the stream as a file
        try:
            response = openai.File.create(file=temp_file, purpose='fine-tune')
        except Exception as e:
            return

        # here can be sure that datasets were read in as that is checked in the finetune_check
        align_dataset_size = self.dataset_sizes[SYMBOLIC_ALIGNMENTS][func_hash] if func_hash in self.dataset_sizes[
            SYMBOLIC_ALIGNMENTS] else 0
        patch_dataset_size = self.dataset_sizes[PATCHES][func_hash] if func_hash in self.dataset_sizes[PATCHES] else 0
        total_dataset_size = align_dataset_size + patch_dataset_size
        training_file_id = response["id"]
        # submit the finetuning job
        try:
            finetuning_response = openai.FineTuningJob.create(training_file=training_file_id, model="gpt-3.5-turbo",
                                                              suffix=finetune_hash)
        except Exception as e:
            return

        self.function_configs[func_hash]["current_training_run"] = {"job_id": finetuning_response["id"],
                                                                    "trained_on_datapoints": total_dataset_size,
                                                                    "last_checked": datetime.datetime.now().strftime(
                                                                        "%Y-%m-%d %H:%M:%S")}
        # update the config json file
        try:
            self._update_config_file(func_hash)
        except Exception as e:
            print(e)
            print("Could not update config file to register a finetuning run")

    def _check_finetuning_status(self, func_hash):
        """
        Check the status of the current finetuning job
        If the job is finished, update the config file to reflect the new model
        """

        job_id = self.function_configs[func_hash]["current_training_run"]["job_id"]
        last_checked = self.function_configs[func_hash]["current_training_run"]["last_checked"]
        # check if last checked was more than 30 mins ago
        if (datetime.datetime.now() - datetime.datetime.strptime(last_checked,
                                                                 "%Y-%m-%d %H:%M:%S")).total_seconds() > 1800:
            response = openai.FineTuningJob.retrieve(job_id)
            self.function_configs[func_hash]["current_training_run"]["last_checked"] = datetime.datetime.now().strftime(
                "%Y-%m-%d %H:%M:%S")
            if response["status"] == "succeeded" or response["status"] == "failed":
                self._update_finetune_config(response, func_hash, response["status"])
            else:
                self._update_config_file(func_hash)

    def _update_finetune_config(self, response, func_hash, status):
        """
        Update the config file to reflect the new model and switch the current model to the finetuned model
        """
        if status == "failed":
            self.function_configs[func_hash]["current_training_run"] = {}
        else:
            self.function_configs[func_hash]["distilled_model"] = response["fine_tuned_model"]
            self.function_configs[func_hash]["last_training_run"] = self.function_configs[func_hash][
                "current_training_run"]
            self.function_configs[func_hash]["current_model_stats"] = {
                "trained_on_datapoints": self.function_configs[func_hash]["current_training_run"][
                    "trained_on_datapoints"],
                "running_faults": []}
            self.function_configs[func_hash]["nr_of_training_runs"] += 1
            self.function_configs[func_hash]["current_training_run"] = {}
        try:
            self._update_config_file(func_hash)
        except Exception as e:
            print(e)
            print("Could not update config file after a successful finetuning run")
            pass



import json
from typing import Any

from tanuki.function_modeler import FunctionModeler
from tanuki.language_models.openai_api import OpenAI_API
from tanuki.models.function_description import FunctionDescription
from tanuki.models.function_example import FunctionExample
from tanuki.models.language_model_output import LanguageModelOutput
from tanuki.utils import approximate_token_count
from tanuki.validator import Validator


class LanguageModelManager(object):
    """
    The LanguageModelManager is responsible for managing the language models and their outputs operationally,
    this includes:
    - Generating outputs from the language models
    - Repairing outputs from the language models
    - Saving outputs from the language models
    - Finetuning the language models from the saved outputs
    """

    def __init__(self, function_modeler: FunctionModeler, generation_token_limit=512) -> None:
        self.function_modeler = function_modeler
        self.instruction = "You are given below a function description and input data. The function description of what the function must carry out can be found in the Function section, with input and output type hints. The input data can be found in Input section. Using the function description, apply the function to the Input and return a valid output type, that is acceptable by the output_class_definition and output_class_hint. Return None if you can't apply the function to the input or if the output is optional and the correct output is None.\nINCREDIBLY IMPORTANT: Only output a JSON-compatible string in the correct response format."
        self.system_message = f"You are a skillful and accurate language model, who applies a described function on input data. Make sure the function is applied accurately and correctly and the outputs follow the output type hints and are valid outputs given the output types."

        self.instruction_token_count = approximate_token_count(self.instruction)
        self.system_message_token_count = approximate_token_count(self.system_message)
        self.api_models = {"openai": OpenAI_API()}
        self.repair_instruction = "Below are an outputs of a function applied to inputs, which failed type validation. The input to the function is brought out in the INPUT section and function description is brought out in the FUNCTION DESCRIPTION section. Your task is to apply the function to the input and return a correct output in the right type. The FAILED EXAMPLES section will show previous outputs of this function applied to the data, which failed type validation and hence are wrong outputs. Using the input and function description output the accurate output following the output_class_definition and output_type_hint attributes of the function description, which define the output type. Make sure the output is an accurate function output and in the correct type. Return None if you can't apply the function to the input or if the output is optional and the correct output is None."
        self.generation_length = generation_token_limit
        self.models = {
            "gpt-4-1106-preview": {
                "token_limit": 128000 - self.generation_length, "type": "openai"
            },
            "gpt-4": {
                "token_limit": 8192 - self.generation_length, "type": "openai"
            },
            "gpt-4-32k": {
                "token_limit": 32768 - self.generation_length, "type": "openai"}
        }  # models and token counts

    def __call__(self,
                 args,
                 function_description: FunctionDescription,
                 kwargs,
                 validator: Validator) -> Any:
        output = self.generate(args, kwargs, function_description)
        # start parsing the object, very hacky way for the time being
        choice_parsed = self._parse_choice(output)
        valid = validator.check_type(choice_parsed, function_description.output_type_hint)
        if not valid:
            choice, choice_parsed, successful_repair = self.repair_output(self,
                                                                          args,
                                                                          kwargs,
                                                                          function_description,
                                                                          output.generated_response,
                                                                          validator)

            if not successful_repair:
                raise TypeError(
                    f"Output type was not valid. Expected an object of type {function_description.output_type_hint}, got '{output.generated_response}'")
            output.generated_response = choice
            output.distilled_model = False
        datapoint = FunctionExample(args, kwargs, output.generated_response)
        if output.suitable_for_finetuning and not output.distilled_model:
            self.function_modeler.postprocess_symbolic_datapoint(function_description.__hash__(), function_description,
                                                                 datapoint, repaired=not valid)
        instantiated = validator.instantiate(choice_parsed, function_description.output_type_hint)
        return instantiated

    def _parse_choice(self, output):
        try:
            # json load
            choice_parsed = json.loads(output.generated_response)
        except:
            # if it fails, it's not a json object, try eval
            try:
                choice_parsed = eval(output.generated_response)
            except:
                choice_parsed = output.generated_response
        return choice_parsed

    def generate(self, args, kwargs, function_description, llm_parameters={}):
        """
        The main generation function, given the args, kwargs, function description and model type, generate a response and check if the datapoint can be saved to the finetune dataset
        """

        prompt, model, save_to_finetune, is_distilled_model = self.get_generation_case(args, kwargs,
                                                                                       function_description)
        if is_distilled_model:
            model_type = self._get_distillation_model_type(model)
        else:
            model_type = self._get_teacher_model_type(model)
        choice = self._synthesise_answer(prompt, model, model_type, llm_parameters)

        output = LanguageModelOutput(choice, save_to_finetune, is_distilled_model)
        return output

    def _synthesise_answer(self, prompt, model, model_type, llm_parameters):
        """
        Synthesise an answer given the prompt, model, model_type and llm_parameters
        """
        if model_type == "openai":
            return self.api_models[model_type].generate(model, self.system_message, prompt, **llm_parameters)
        else:
            raise NotImplementedError("Only OpenAI is supported currently. " + \
                                      "Please feel free to raise a PR to support development")

    def _get_distillation_model_type(self, model):
        """
        Get the distilled model type given the model
        """
        # currently only openai is supported
        return "openai"

    def _get_teacher_model_type(self, model):
        """
        Get the teacher model type given the model
        """
        # check if model is in the models
        if model in self.models.keys():
            return self.models[model]["type"]
        else:
            raise ValueError("This teacher model is not supported")

    def get_generation_case(self, args, kwargs, function_description):
        """
        Get the generation case with the correct prompt and model
        First get the current model, then if distilled model, do zero-shot prompt and return False as suitable_for_finetune
        If not distilled model, check if suitable for finetuning, create the prompt and return the correct model given the token count
        """
        f = str(function_description.__dict__.__repr__())

        distilled_model, teacher_models = self.function_modeler.get_models(function_description)
        is_distilled_model = distilled_model != ""
        suitable_for_distillation, input_prompt_token_count = self.suitable_for_finetuning_token_check(args, kwargs, f,
                                                                                                       self.function_modeler.distillation_token_limit)
        # no examples needed, using a finetuned model. Dont save to finetune dataset
        if is_distilled_model and suitable_for_distillation:
            prompt = self.construct_prompt(f, args, kwargs, None)
            return prompt, distilled_model, suitable_for_distillation, True

        else:
            aligns = self.function_modeler.get_symbolic_alignments(function_description.__hash__(), max=16)
            examples = "\n".join(
                [f"Inputs:\nArgs: {align['args']}\nKwargs: {align['kwargs']}\nOutput: {align['output']}" for align in
                 aligns])
            prompt = self.construct_prompt(f, args, kwargs, examples)
            examples_token_count = approximate_token_count(examples)
            total_token_count = examples_token_count + input_prompt_token_count + self.instruction_token_count + self.system_message_token_count
            model = self.choose_model_from_tokens(teacher_models, total_token_count)
            if model:
                return prompt, model, suitable_for_distillation, False
            else:
                raise ValueError(
                    "The input content and align statements combined are too long, please shorten it. The maximum currently allowed token limit is 32000")

    def suitable_for_finetuning_token_check(self, args, kwargs, f, distillation_token_count):
        """
        Check if the inputs are suitable for finetuning, i.e are below the finetuning token count
        """
        # check if finetunable
        finetuning_prompt = f"Function: {f}\n---\nInputs:\nArgs: {args}\nKwargs: {kwargs}\nOutput:"
        input_prompt_token_count = approximate_token_count(finetuning_prompt)
        suitable_for_finetune = input_prompt_token_count + self.instruction_token_count + self.system_message_token_count < distillation_token_count
        return suitable_for_finetune, input_prompt_token_count

    def construct_prompt(self, f, args, kwargs, examples):
        """
        Cosntruct a prompt given the function description, args, kwargs and examples
        """
        example_input = f"Examples:{examples}\n" if examples else ""
        content = f"{self.instruction}\nFunction: {f}\n{example_input}---\nInputs:\nArgs: {args}\nKwargs: {kwargs}\nOutput:"
        return content

    def repair_generate(self, args, kwargs, f, failed_outputs_list, examples, models):
        """
        Repair the output given the input, function description, failed outputs list, examples and models
        """
        prompt = self.generate_repair_prompt(args, kwargs, f, failed_outputs_list, examples)
        prompt_token_count = approximate_token_count(prompt)
        model = self.choose_model_from_tokens(models, prompt_token_count)
        if model:
            model_type = self._get_teacher_model_type(model)
            choice = self._synthesise_answer(prompt, model, model_type, {})
            return choice
        else:
            return None

    def generate_repair_prompt(self, args, kwargs, f, failed_outputs_list, examples):
        """
        Generate a repair prompt given the args, kwargs, function description, failed outputs list and examples
        """

        failed_examples = ""
        for failed_output in failed_outputs_list:
            failed_examples += f"Output: {failed_output[0]}\nError: {failed_output[1]}\n\n"
        successful_examples = f"Successful Examples:{examples}\n" if examples else ""
        prompt = f"{self.repair_instruction}\nFUNCTION DESCRIPTION: {f}\n{successful_examples}---Inputs:\nArgs: {args}\nKwargs: {kwargs}\nFAILED EXAMPLES: {failed_examples}Correct output:"
        return prompt

    def choose_model_from_tokens(self, models, token_count):
        """
        Choose a model from the models given the token count
        """

        for model in models:
            # check if model is in the models
            if model in self.models.keys():
                if token_count < self.models[model]["token_limit"]:
                    return model
        return None

    def repair_output(self,
                      args: tuple,
                      kwargs: dict,
                      function_description: FunctionDescription,
                      choice,
                      validator: Validator) -> tuple:
        """
        Repair an output, that failed type validation by generating a new output using the teacher model and the error
        Args:
            args (tuple): The args of the function
            kwargs (dict): The kwargs of the function
            function_description (FunctionDescription): The function description
            choice: The output that failed type validation, type is arbitrary
            validator (Validator): The validator object

        Returns:
            choice (str): The choice that was generated by the language model
            choice_parsed: The parsed choice, type is arbitrary
            valid (bool): Whether the output was correctly repaired was valid
        """

        # get the teacher models
        teacher_models = self.function_modeler.get_models(function_description)[1]
        valid = False
        retry_index = 5
        f = str(function_description.__dict__.__repr__() + "\n")
        error = f"Output type was not valid. Expected an valid object of type {function_description.output_type_hint}, got '{choice}'"
        # instantiate the failed outputs list
        failed_outputs_list = [(choice, error)]
        while retry_index > 0 and not valid:
            # get the alignments
            aligns = self.function_modeler.get_symbolic_alignments(function_description.__hash__(), max=5)
            examples = "\n".join(
                [f"Inputs:\nArgs: {align['args']}\nKwargs {align['kwargs']}\nOutput: {align['output']}" for align in
                 aligns])
            # Generate the reparied LLM output
            choice = self.repair_generate(args, kwargs, f, failed_outputs_list, examples, teacher_models)
            if not choice:
                # if no choice then the input was too long for the model
                # no specific error but the retry index goes down
                retry_index -= 1
                continue

            # start parsing the object
            try:
                # json load
                choice_parsed = json.loads(choice)
            except:
                # if it fails, it's not a json object, try eval
                try:
                    choice_parsed = eval(choice)
                except:
                    choice_parsed = choice

            valid = validator.check_type(choice_parsed, function_description.output_type_hint)
            if not valid:
                # if it's not valid, add it to the failed outputs list
                error = f"Output type was not valid. Expected an object of type {function_description.output_type_hint}, got '{choice}'"
                failed_outputs_list.append((choice, error))
                retry_index -= 1

        return choice, choice_parsed, valid



import os

from bitarray._bitarray import bitarray

from tanuki.persistence.filter.bloom_interface import IBloomFilterPersistence


class BloomFilterFileSystemDriver(IBloomFilterPersistence):
    """
    This is a Filesystem implementation of a Bloom Filter persistence layer.
    """

    def __init__(self, log_directory: str):
        self.log_directory = log_directory

    def save(self, bit_array: bitarray) -> None:
        """
        Write a bloom filter array of bits to the local filesystem.
        :param bloom_filter: A bloom filter which tracks unique function invocations
        """
        bloom_filter_path = os.path.join(self.log_directory, 'bloom_filter_state.bin')
        # Append 0 bits to make the length a multiple of 8
        while len(bit_array) % 8 != 0:
            bit_array.append(0)

        with open(bloom_filter_path, 'wb') as f:
            f.write(bit_array.tobytes())

    def load(self) -> bitarray:
        """
        Load a bloom filter from the local filesystem.
        :return: A bloom filter object containing the state of unique function invocations
        """
        bloom_filter_path = os.path.join(self.log_directory, 'bloom_filter_state.bin')
        with open(bloom_filter_path, 'rb') as f:
            bit_array = bitarray()
            bit_array.frombytes(f.read())

        while len(bit_array) % 8 != 0:
            bit_array.append(0)

        return bit_array


from tanuki.validator import Validator


def test_base_type():
    print("test_base_type")
    validator = Validator()
    assert validator.is_base_type(int)
    assert validator.is_base_type(float)
    assert validator.is_base_type(str)
    assert validator.is_base_type(bool)
    assert validator.is_base_type(None)
    assert not validator.is_base_type(list)
    assert not validator.is_base_type(dict)
    assert not validator.is_base_type(tuple)
    assert not validator.is_base_type(set)


if __name__ == "__main__":
    test_base_type()


from typing import List
from tanuki.register import Register

import os
from typing import Optional, Literal, List
import openai
from dotenv import load_dotenv
import tanuki

load_dotenv()
openai.api_key = os.getenv("OPENAI_API_KEY")


@tanuki.patch
def classify_sentiment_2(input: str, input_2: str) -> Optional[Literal['Good', 'Bad']]:
    """
    Determine if the inputs are positive or negative sentiment, or None
    """


@tanuki.patch(environment_id = 12, ignore_finetune_fetching=True, ignore_finetuning=True, ignore_data_storage=True)
def classify_sentiment(input: str) -> Optional[Literal['Good', 'Bad']]:
    """
    Determine if the input is positive or negative sentiment
    """

@tanuki.align
def align_classify_sentiment():
    """We can test the function as normal using Pytest or Unittest"""

    i_love_you = "I love you"
    assert classify_sentiment_2(i_love_you, "I love woo") == 'Good'
    assert classify_sentiment_2("I hate you", "You're discusting") == 'Bad'
    assert classify_sentiment_2("Today is wednesday", "The dogs are running outside") == None


    assert classify_sentiment("I love you") == 'Good'
    assert classify_sentiment("I hate you") == 'Bad'
    assert classify_sentiment("Wednesdays are in the middle of the week") == None

def test_classify_sentiment():
    align_classify_sentiment()
    bad_input = "I find you awful"
    good_input = "I really really like you"
    good_input_2 = "I adore you"
    assert classify_sentiment("I like you") == 'Good'
    assert classify_sentiment(bad_input) == 'Bad'
    assert classify_sentiment("I am neutral") == None

    assert classify_sentiment_2(good_input, good_input_2) == 'Good'
    assert classify_sentiment_2("I do not like you you", bad_input) == 'Bad'
    assert classify_sentiment_2("I am neutral", "I am neutral too") == None

def test_configurability():
    classify_sent_description = Register.load_function_description(classify_sentiment)
    classify_sentiment_2_description = Register.load_function_description(classify_sentiment_2)
    sent_func_hash = classify_sent_description.__hash__()
    sent_func_2_hash = classify_sentiment_2_description.__hash__()

    func_modeler = tanuki.function_modeler
    assert func_modeler.environment_id == 12
    assert sent_func_hash in func_modeler.check_finetune_blacklist
    assert sent_func_2_hash not in func_modeler.check_finetune_blacklist
    assert sent_func_hash in func_modeler.execute_finetune_blacklist
    assert sent_func_2_hash not in func_modeler.execute_finetune_blacklist
    assert sent_func_hash in func_modeler.store_data_blacklist
    assert sent_func_2_hash not in func_modeler.store_data_blacklist







import ast
import builtins
import sys
import uuid
from _ast import NotEq, NotIn, IsNot, In
from typing import Optional, List, Dict, Callable

from tanuki.utils import get_key


class Or(list):
    pass


class AssertionVisitor(ast.NodeVisitor):
    def __init__(self,
                 scope: Optional[dict] = None,
                 patch_symbolic_funcs: Dict[str, Callable] = {},
                 patch_embeddable_funcs: Dict[str, Callable] = {},
                 wrapper_alias='test_func'):

        # This is for storing positive asserts for both embeddable and symbolic functions, i.e where the outputs should
        # be the same.
        self.mocks = {}  # {args: output}

        # This is for storing negative asserts for embeddable functions, i.e functions that should be embedded
        # away from each other in the latent space.
        self.negative_mocks = {}  # {args: args}

        self.scopes = [{}]  # Stack of scopes to mimic variable scope in code
        self.imported_modules = {}  # keys are module names, values are the actual modules
        self.patch_symbolic_names = list(patch_symbolic_funcs.keys())  # names of symbolic functions to patch
        self.patch_symbolic_funcs = patch_symbolic_funcs  # symbolic functions to patch
        self.patch_embeddable_names = list(patch_embeddable_funcs.keys())  # names of embeddable functions to patch
        self.patch_embeddable_funcs = patch_embeddable_funcs  # embeddable functions to patch

        if scope:
            self.local_scope = scope

        current_module = sys.modules[__name__]
        self.imported_modules[current_module.__name__] = current_module
        self.wrapper_alias = wrapper_alias

    def load_variable_values(self, var_name):
        for scope in reversed(self.scopes):
            if var_name in scope:
                return scope[var_name]
        return None

    def visit_Import(self, node):
        for name in node.names:
            self.imported_modules[name.name.split('.')[0]] = __import__(name.name)

    def visit_ImportFrom(self, node):
        module_name = node.module
        self.imported_modules[module_name.split('.')[0]] = __import__(module_name)

    def visit_Assign(self, node):
        target = node.targets[0]
        value = node.value
        if isinstance(target, ast.Name):
            if isinstance(value, ast.List):
                self.scopes[-1][target.id] = [self.get_value(elt) for elt in value.elts]
            elif isinstance(value, ast.Name):
                self.scopes[-1][target.id] = self.load_variable_values(value.id)
            else:
                self.scopes[-1][target.id] = self.get_value(value)

    def get_value(self, node):
        if isinstance(node, ast.Str):
            return node.s
        elif isinstance(node, ast.Num):
            return node.n
        elif isinstance(node, ast.NameConstant):
            return node.value
        elif isinstance(node, ast.List):
            return [self.get_value(e) for e in node.elts]
        elif isinstance(node, ast.Dict):
            keys = [self.get_value(k) for k in node.keys]
            values = [self.get_value(v) for v in node.values]
            return dict(zip(keys, values))
        elif isinstance(node, ast.Name):
            value = self.load_variable_values(node.id)
            # check if value has a __dict__ attribute
            if hasattr(value, "__dict__"):
                return value.__dict__
            else:
                return value
        elif isinstance(node, ast.Call):
            return self.extract_output(node)
        else:
            # This is a simplification; you might want to add more types
            raise NotImplementedError(f"Node type {type(node)} not supported")

    def process_assert(self, node, iter_name=None, evaluated_expr=None):

        if isinstance(node.test, ast.UnaryOp) and isinstance(node.test.op, ast.Not):
            # Special case for 'assert not ...'. We have to convert it into 'assert ... == None'
            operand = node.test.operand
            self.process_assert_helper(operand, ast.NameConstant(value=None), iter_name, op=ast.Eq())
            return

        left = node.test.left
        right = node.test.comparators[0]

        if iter_name:
            for input_val in evaluated_expr:
                self.scopes.append({iter_name: input_val})
                self.process_assert_helper(left, right, iter_name, op=node.test.ops[0])
                self.scopes.pop()
        else:
            self.process_assert_helper(left, right, op=node.test.ops[0])

    def process_assert_with_tuple(self, node, iter_names, evaluated_expr):
        left = node.test.left
        right = node.test.comparators[0]

        if isinstance(node.test.ops[0], (NotEq, NotIn, IsNot)):
            return

        if evaluated_expr is not None:
            # Loop through the evaluated expression (which should be an iterable)
            for input_tuple in evaluated_expr:
                # Create a new scope for these variables
                self.scopes.append(dict(zip(iter_names, input_tuple)))

                if isinstance(node.test, ast.UnaryOp) and isinstance(node.test.op, ast.Not):
                    # Special case for 'assert not ...'
                    #operand = node.test.operand
                    self.process_assert_helper(left, ast.NameConstant(value=None), op=ast.Eq())
                else:
                    # Process the assert statement
                    self.process_assert_helper(left, right, op=node.test.ops[0])

                # Remove the temporary scope
                self.scopes.pop()

    def process_assert_helper(self, left, right, iter_name=None, op=None):
        """
        This is a helper function for processing asserts. It is low-level and is called by higher-level functions
        that analyze the AST of the aligned function.

        It handles the case where both sides of the assert statement are patched embedding functions: where we need to
        do special mocking.

        It also handles the case where the left side of the assert statement is a patched symbolic function.

        Args:
            left: The expression on the left of an assert statement
            right: The expression on the right of an assert statement
            iter_name: The name of the iterator variable (if any)
            op: The operator used in the assert statement (e.g. ast.Eq() for '==')
        """
        if self.is_embeddable_function_call(left) and self.is_embeddable_function_call(right):
            # Both sides are patched embeddable functions
            self.process_assert_helper_both_sides_embeddable(left, right, iter_name, op)
        elif self.is_embeddable_function_call(left) and not self.is_embeddable_function_call(right):
            raise ValueError("Cannot compare patched embeddable function with non-patched embeddable function")
        elif not self.is_embeddable_function_call(left) and self.is_embeddable_function_call(right):
            raise ValueError("Cannot compare non-patched embeddable function with patched embeddable function")
        else:
            # Only equality operators make sense when dealing with symbolic functions.
            # e.g Telling an LLM that a function should not yield X is not meaningful
            # (as the set of possible outputs is infinite)
            if isinstance(op, (NotEq, NotIn, IsNot)):
                return

            if isinstance(left, ast.Call):
                if isinstance(left.func, ast.Name) and left.func.id in self.patch_symbolic_names:
                    self.process_assert_helper_lr(left, right, iter_name, op)
                elif isinstance(left.func, ast.Attribute) and left.func.attr in self.patch_symbolic_names:
                    self.process_assert_helper_lr(left, right, iter_name, op)
                else:
                    if isinstance(right.func, ast.Name) and right.func.id in self.patch_symbolic_names:
                        self.process_assert_helper_lr(right, left, iter_name, op)
                    elif isinstance(right.func, ast.Attribute) and right.func.attr in self.patch_symbolic_names:
                        self.process_assert_helper_lr(right, left, iter_name, op)

            elif isinstance(right, ast.Call):
                if isinstance(right.func, ast.Name) and right.func.id in self.patch_symbolic_names:
                    self.process_assert_helper_lr(right, left, iter_name, op)
                elif isinstance(right.func, ast.Attribute) and right.func.attr in self.patch_symbolic_names:
                    self.process_assert_helper_lr(right, left, iter_name, op)

    def process_assert_helper_both_sides_embeddable(self, left, right, iter_name=None, op=None):

        if hasattr(left.func, 'attr') and hasattr(right.func, 'attr') and left.func.attr != right.func.attr:
            raise ValueError(f"Cannot compare two different patched embeddable functions: {left} and {right}")
        if hasattr(left.func, 'id') and hasattr(right.func, 'id') and left.func.id != right.func.id:
            raise ValueError(f"Cannot compare two different patched embeddable functions: {left} and {right}")

        left_args, left_kwargs = self.extract_args(left, iter_name)
        right_args, right_kwargs = self.extract_args(right, iter_name)

        left_key = get_key(left_args, left_kwargs)
        right_key = get_key(right_args, right_kwargs)

        if isinstance(op, ast.Eq):
            # For equality, both sides should produce the same mock value
            if left_key in self.mocks:
                self.mocks[right_key] = self.mocks[left_key]
            elif right_key in self.mocks:
                self.mocks[left_key] = self.mocks[right_key]
            else:
                mock_value = self.generate_mock_embedding()
                self.mocks[left_key] = mock_value
                self.mocks[right_key] = mock_value
        elif isinstance(op, ast.NotEq):
            # For inequality, ensure different mock values
            if left_key in self.mocks:
                self.mocks[right_key] = self.generate_mock_embedding()
            elif right_key in self.mocks:
                self.mocks[left_key] = self.generate_mock_embedding()
            else:
                self.mocks[left_key] = self.generate_mock_embedding()
                self.mocks[right_key] = self.generate_mock_embedding()

            self.negative_mocks[left_key] = right_key

    def generate_mock_embedding(self):
        # Method to generate a unique mock value
        return f"MockEmbedding_{uuid.uuid4()}"

    def is_embeddable_function_call(self, node):
        # Helper method to determine if a node is a call to a patched function
        if isinstance(node, ast.Call):
            if isinstance(node.func, ast.Name) and node.func.id in self.patch_embeddable_funcs:
                return True
            elif isinstance(node.func, ast.Attribute) and node.func.attr in self.patch_embeddable_funcs:
                return True
        return False

    def process_assert_helper_lr(self, left, right, iter_name=None, op=None):
        input_args, input_kwargs = self.extract_args(left, iter_name)
        if isinstance(op, In):
            output = Or(self.extract_output(right))
        else:
            output = self.extract_output(right)

        key = get_key(input_args, input_kwargs)

        self.mocks[key] = output

    def extract_args(self, node, iter_name=None):
        # Assuming all arguments are positional
        args = []
        current_scope = self.scopes[-1]
        for arg in node.args:
            value = self.eval_expr(arg)
            if callable(value):
                if iter_name is not None:
                    arg_values = [current_scope.get(arg_name, None) for arg_name in iter_name]
                    value = value(*arg_values)
                else:
                    value = value()
            elif value is None:
                # If evaluation returns None, fall back to scope
                value = current_scope.get(arg.id, None)
            args.append(value)

        # Extract keyword arguments
        kwargs = {}
        for kw in node.keywords:
            kw_value = self.eval_expr(kw.value)
            if callable(kw_value):
                if iter_name is not None:
                    kw_values = [current_scope.get(arg_name, None) for arg_name in iter_name]
                    kw_value = kw_value(*kw_values)
                else:
                    kw_value = kw_value()
            elif kw_value is None:
                # If evaluation returns None, fall back to scope
                kw_value = current_scope.get(kw.arg, None)
            kwargs[kw.arg] = kw_value

        return args, kwargs

    def eval_expr(self, node):
        current_scope = self.scopes[-1]
        if isinstance(node, ast.Name):
            # Prioritize variables in the current scope over built-in functions
            if node.id in current_scope:
                return current_scope[node.id]
            elif hasattr(builtins, node.id):
                func = getattr(builtins, node.id)
                return func

        elif isinstance(node, ast.Call):
            # Function call, could be 'zip' or other
            func = self.eval_expr(node.func)  # Recursively evaluate the function expression
            if func is None:
                raise ValueError(
                    f"Function {node.func.id if isinstance(node.func, ast.Name) else ''} not found in the current scope or built-ins.")

            args = [self.eval_expr(arg) for arg in node.args]  # Recursively evaluate the arguments
            return func(*args)  # Assume func_name is a Python built-in, e.g., zip
        elif isinstance(node, ast.Num):
            return node.n  # Assume it's a number for simplicity
        elif isinstance(node, ast.Str):
            return node.s  # Assume it's a string for simplicity
        elif isinstance(node, ast.List):
            return [self.eval_expr(x) for x in node.elts]  # Assume it's a list for simplicity
        elif isinstance(node, ast.Tuple):
            return tuple([self.eval_expr(x) for x in node.elts])  # Assume it's a tuple for simplicity
        elif isinstance(node, ast.Dict):
            return {self.eval_expr(k): self.eval_expr(v) for k, v in zip(node.keys, node.values)}
        else:
            raise NotImplementedError(f"Node type {type(node).__name__} not handled yet")

    def instantiate(self, func, *args, **kwargs):
        """Instantiate a function with the given arguments and keyword arguments."""
        try:
            return func(*args, **kwargs)
        except TypeError as e:
            # If the function cannot be instantiated, return the function itself
            return func
        except Exception as e:
            raise e

    def extract_output(self, node, scope=None):
        # current_scope = self.scopes[-1]

        def eval_args(args, scope):
            return [self.extract_output(arg, scope) for arg in args]

        def eval_kwargs(keywords, scope):
            return {kw.arg: self.extract_output(kw.value, scope) for kw in keywords}

        if isinstance(node, ast.Num):
            return node.n
        elif isinstance(node, ast.Str):
            return node.s
        elif isinstance(node, ast.List):
            return [self.extract_output(elt, scope) for elt in node.elts]
        elif isinstance(node, ast.Dict):
            return {self.extract_output(k, scope): self.extract_output(v, scope) for k, v in
                    zip(node.keys, node.values)}
        elif isinstance(node, ast.Name):
            return self.load_variable_values(node.id)
        elif isinstance(node, ast.Call):
            if isinstance(node.func, ast.Name):
                func_name = node.func.id
                args = eval_args(node.args, scope)
                kwargs = eval_kwargs(node.keywords, scope)

                # if func_name == "TodoItem":
                #     return TodoItem(*args, **kwargs)
                # elif func_name == "datetime":
                #     return datetime.datetime(*args, **kwargs)
                _globals = globals()
                _function_globals = self.local_scope[self.wrapper_alias].__globals__ if self.wrapper_alias in self.local_scope else {}
                # Generalized object instantiation
                if func_name in self.imported_modules:
                    module = self.imported_modules[func_name]
                    if hasattr(module, func_name):
                        func = getattr(module, func_name)
                        return self.instantiate(func, *args, **kwargs)

                elif func_name in _globals:
                    func = _globals[func_name]
                    return self.instantiate(func, *args, **kwargs)
                elif func_name in self.local_scope:
                    func = self.local_scope[func_name]
                    return self.instantiate(func, *args, **kwargs)
                elif func_name in _function_globals:
                    func = _function_globals[func_name]
                    return self.instantiate(func, *args, **kwargs)
                else:
                    raise NotImplementedError(f"Function {func_name} not handled yet")
            elif isinstance(node.func, ast.Attribute):
                func_name = node.func.attr
                # obj = self.extract_output(node.func.value, scope)
                _globals = globals()
                _function_globals = self.local_scope[self.wrapper_alias].__globals__ if self.wrapper_alias in self.local_scope else {}
                if isinstance(node.func.value, ast.Name):
                    module_or_class_name = node.func.value.id

                    if module_or_class_name == 'self':
                        # Handling methods called on 'self'
                        args = eval_args(node.args, scope)
                        kwargs = eval_kwargs(node.keywords, scope)
                        patched_func_scope = self.patch_symbolic_funcs[func_name]
                        #method = getattr(scope['self'], func_name, None)
                        if patched_func_scope:
                            return self.instantiate(patched_func_scope, *args, **kwargs)
                        else:
                            raise NotImplementedError(f"Method {func_name} not found in class")

                    # Check if module_or_class_name is an imported module
                    if module_or_class_name in self.imported_modules:
                        obj = self.imported_modules[module_or_class_name]
                    # Check if module_or_class_name is a globally defined class
                    elif module_or_class_name in _globals:
                        obj = _globals[module_or_class_name]
                    elif module_or_class_name in builtins.__dict__:
                        obj = builtins.__dict__[module_or_class_name]
                    elif module_or_class_name in self.local_scope:
                        obj = self.local_scope[module_or_class_name]
                    elif module_or_class_name in _function_globals:
                        obj = _function_globals[module_or_class_name]
                    else:
                        raise NotImplementedError(f"Module or class {module_or_class_name} not found")
                else:
                    obj = self.extract_output(node.func.value, scope)
                args = eval_args(node.args, scope)
                kwargs = eval_kwargs(node.keywords, scope)

                if hasattr(obj, func_name):
                    func = getattr(obj, func_name)
                    return self.instantiate(func, *args, **kwargs)
                else:
                    raise NotImplementedError(f"Attribute-based call {obj}.{func_name} not handled yet")

            # Add additional function handling here
        elif isinstance(node, ast.Attribute):
            # Handle attributes like datetime.datetime
            obj = self.extract_output(node.value, scope)
            return getattr(obj, node.attr)
        elif isinstance(node, ast.Tuple):
            return tuple(self.extract_output(elt, scope) for elt in node.elts)
        elif isinstance(node, ast.Constant):
            return node.value

        raise NotImplementedError(f"Node type {type(node).__name__} not handled yet")

    def visit_For(self, node):
        if isinstance(node.target, ast.Name):
            evaluated_expr = self.eval_expr(node.iter)
            iter_name = node.target.id
            if iter_name:
                for stmt in node.body:
                    if isinstance(stmt, ast.Assert):
                        self.process_assert(stmt, iter_name, evaluated_expr)
        # If loop variable is a tuple
        elif isinstance(node.target, ast.Tuple):
            iter_names = [elt.id for elt in node.target.elts]  # Extract names from tuple
            # iter_name = self.extract_variable_name(node.iter)
            evaluated_expr = self.eval_expr(node.iter)
            for stmt in node.body:
                if isinstance(stmt, ast.Assert):
                    self.process_assert_with_tuple(stmt, iter_names, evaluated_expr)

        self.scopes.pop()

    def extract_variable_name(self, node):
        if isinstance(node, ast.Name):
            return node.id
        elif isinstance(node, ast.Call):
            if isinstance(node.func, ast.Name):
                return node.func.id  # For now, just return the function name. You may also traverse args.
        return None  # Or raise an exception depending on your need

    def visit_Assert(self, node):
        self.process_assert(node)

    def visit_FunctionDef(self, node):
        self.scopes.append({})
        self.generic_visit(node)
        self.scopes.pop()



import os
import unittest
from typing import List
from unittest import TestCase

import numpy as np
import openai
from dotenv import load_dotenv
from tanuki.models.embedding import Embedding
import tanuki

load_dotenv()
openai.api_key = os.getenv("OPENAI_API_KEY")


class TestEmbedding(TestCase):

    @tanuki.patch
    def embed_sentiment(self, input: str) -> Embedding[np.ndarray]:
        """
        Determine if the inputs are positive or negative sentiment, or None
        """

    @tanuki.patch
    def embed_sentiment2(self, input: str) -> Embedding[np.ndarray]:
        """
        Determine if the inputs are positive or negative sentiment, or None
        """

    @tanuki.patch
    def is_positive_sentiment(self, input: str) -> bool:
        """
        Determine if the inputs are positive or negative sentiment
        """

    @tanuki.align
    def align_embed_sentiment(self) -> None:
        """
        Align some embed_sentiment functions
        """
        # We try to push these embeddings apart
        assert self.embed_sentiment("I love this movie") != self.embed_sentiment("I hate this movie")

        # And push these embeddings together
        assert self.embed_sentiment("I love this movie") == self.embed_sentiment("I love this film")
        assert self.embed_sentiment("I love this movie") == self.embed_sentiment("I loved watching the movie")

    @tanuki.align
    def broken_heterogenous_align(self) -> None:
        # This should break, because the embedding function 'embed_sentiment' must be compared to itself with
        # different inputs.
        # TODO: We should investigate supporting alignment between heterogenous functions.
        #  Will functions be trained piecewise?
        assert self.embed_sentiment("I love this movie") == self.embed_sentiment2("I hate this movie")

    @tanuki.align
    def broken_align_embed_sentiment(self) -> None:
        # This should break, because the embedding function 'embed_sentiment' must be compared to another embedding
        # function.
        assert self.embed_sentiment("I love this movie") == "I hate this movie"

    @tanuki.align
    def broken_align_symbolic_with_embeddable(self) -> None:
        # This should break, because the embedding function 'embed_sentiment' must be compared to another embedding
        # function.
        assert self.embed_sentiment("I love this movie") == self.is_positive_sentiment("I hate this movie")

    def test_data_type(self):
        # Test with np.ndarray
        embedding_array = Embedding[np.ndarray]([0, 2, 4])
        self.assertIsInstance(embedding_array._data, np.ndarray)

        # Test with list[float]
        embedding_list = Embedding[List[float]]([0.0, 2.0, 4.0])
        self.assertIsInstance(embedding_list._data, list)
        self.assertTrue(all(isinstance(item, float) for item in embedding_list._data))

    def test_instantiate_np_embedding(self):
        embedding = Embedding[np.ndarray]([0, 2, 4])
        trans = embedding.T
        assert isinstance(trans, np.ndarray)

    def test_get_embedding(self):
        embedding = self.embed_sentiment("I love this movie")
        transposed = embedding.T
        assert isinstance(embedding, Embedding)
        assert isinstance(transposed, np.ndarray)

    def test_align(self):
        self.align_embed_sentiment()

    def test_cannot_align_embeddable_with_literal(self):
        with self.assertRaises(ValueError):
            self.broken_align_embed_sentiment()

    def test_cannot_align_symbolic_with_embeddable(self):
        with self.assertRaises(ValueError):
            self.broken_align_symbolic_with_embeddable()

    def test_cannot_align_heterogenous(self):
        with self.assertRaises(ValueError):
            self.broken_heterogenous_align()

if __name__ == '__main__':
    unittest.main()


import inspect
from typing import get_type_hints, Literal, get_origin, Tuple, Callable, Optional, Dict

from tanuki.models.embedding import Embedding
from tanuki.models.function_description import FunctionDescription
from tanuki.models.function_type import FunctionType

alignable_symbolic_functions = {}
alignable_embedding_functions = {}


class Register:

    def __init__(self):
        pass

    @staticmethod
    def get(func_name) -> Tuple[FunctionType, Callable]:
        if func_name not in alignable_symbolic_functions and func_name not in alignable_embedding_functions:
            pass

        if func_name in alignable_symbolic_functions:
            return FunctionType.SYMBOLIC, alignable_symbolic_functions[func_name]
        elif func_name in alignable_embedding_functions:
            return FunctionType.EMBEDDABLE, alignable_embedding_functions[func_name]

    @staticmethod
    def function_names_to_patch(*args, type: Optional[FunctionType] = None):
        """
        Get the registered function names that should be patched, either globally (if len(args)==0) or as members of
        an instance
        :param args: Optional instance to check
        :return:
        """
        function_names = []
        if len(args) == 1:
            instance = args[0]

            if type == FunctionType.SYMBOLIC:
                for key in alignable_symbolic_functions.keys():
                    if hasattr(instance, key):
                        function_names.append(key)
                return function_names
            elif type == FunctionType.EMBEDDABLE:
                for key in alignable_embedding_functions.keys():
                    if hasattr(instance, key):
                        function_names.append(key)
                return function_names
            else:
                for key in alignable_symbolic_functions.keys():
                    if hasattr(instance, key):
                        function_names.append(key)
                for key in alignable_embedding_functions.keys():
                    if hasattr(instance, key):
                        function_names.append(key)
                return function_names
        else:
            if type == FunctionType.SYMBOLIC:
                return list(alignable_symbolic_functions.keys())
            elif type == FunctionType.EMBEDDABLE:
                return list(alignable_embedding_functions.keys())
            else:
                return list(alignable_symbolic_functions.keys()) + list(alignable_embedding_functions.keys())

    @staticmethod
    def functions_to_patch(*args, type: Optional[FunctionType] = None) -> Dict[str, Callable]:
        function_names = Register.function_names_to_patch(*args, type=type)
        if type == FunctionType.SYMBOLIC:
            return {key: alignable_symbolic_functions[key] for key in function_names}
        elif type == FunctionType.EMBEDDABLE:
            return {key: alignable_embedding_functions[key] for key in function_names}
        else:
            return {key: alignable_symbolic_functions[key] for key in function_names} + \
                   {key: alignable_embedding_functions[key] for key in function_names}

    @staticmethod
    def add_function(func, function_description: FunctionDescription):
        if function_description.type == FunctionType.SYMBOLIC:
            alignable_symbolic_functions[func.__name__] = func
        elif function_description.type == FunctionType.EMBEDDABLE:
            alignable_embedding_functions[func.__name__] = func

    @staticmethod
    def load_function_description_from_name(*args) -> FunctionDescription:
        """
        Load a function description from a function name from the global scope.
        :param func_name:
        :return:
        """
        if len(args) == 1:
            instance = None
            func_name = args[0]
        elif len(args) == 2:
            instance = args[0]
            func_name = args[1]
        else:
            raise ValueError("Invalid number of arguments")

        if not instance:
            if func_name in alignable_symbolic_functions:
                func_object = alignable_symbolic_functions[func_name]
            elif func_name in alignable_embedding_functions:
                func_object = alignable_embedding_functions[func_name]
        else:
            func_object = getattr(instance, func_name)

        return Register.load_function_description(func_object)

    @staticmethod
    def load_function_description(func_object) -> FunctionDescription:
        """
        Create a function description from a function object that can be used to register the function.
        :param func_object:
        :return:
        """
        sig = inspect.signature(func_object)
        type_hints = get_type_hints(func_object)

        # Extract input type hints and output type hint
        input_type_hints = {k: v for k, v in type_hints.items() if k in sig.parameters}
        output_type_hint = type_hints.get('return')

        # Fetch the docstring
        docstring = func_object.__doc__.strip() if func_object.__doc__ else ""

        def get_class_definition(class_type):
            """Helper function to get class definition source if not a built-in type"""
            if hasattr(class_type, "__origin__"):  # Check if it's a generic type
                origin_type = class_type.__origin__
                if origin_type is Literal:  # Handle Literal case
                    return [literal for literal in class_type.__args__]
                elif hasattr(class_type, "__args__"):  # Access inner types
                    return [get_class_definition(arg) for arg in class_type.__args__ if arg is not None]
            elif inspect.isclass(class_type) and class_type.__module__ != "builtins":
                return inspect.getsource(class_type)
            return class_type.__name__

        # Extract class definitions for input and output types
        input_class_definitions = {
            param_name: get_class_definition(param_type)
            for param_name, param_type in input_type_hints.items()
        }
        # if inspect.isclass(output_type_hint) and issubclass(output_type_hint, Embedding):
        #     output_class_definition = None
        # else:
        #     output_class_definition = get_class_definition(output_type_hint)
        output_class_definition = None
        function_type = FunctionType.SYMBOLIC
        if inspect.isclass(output_type_hint):
            # Check if the base class of the output type hint is Embedding
            base_class = get_origin(output_type_hint) or output_type_hint
            if issubclass(base_class, Embedding):
                output_class_definition = None
                function_type = FunctionType.EMBEDDABLE
            else:
                output_class_definition = get_class_definition(output_type_hint)

        return FunctionDescription(
            name=func_object.__name__,
            docstring=docstring,
            input_type_hints=input_type_hints,
            output_type_hint=output_type_hint,
            input_class_definitions=input_class_definitions,
            output_class_definition=output_class_definition,
            type=function_type
        )



import dataclasses
import datetime
import json
import typing
from typing import get_args, Literal
import string
import types

def json_default(thing):
    try:
        return dataclasses.asdict(thing)
    except TypeError:
        pass
    if isinstance(thing, datetime.datetime):
        return thing.isoformat(timespec='microseconds')
    if isinstance(thing, type):
        return thing.__name__
    #if hasattr(typing, "_GenericAlias") and isinstance(thing, typing._GenericAlias):
    if hasattr(typing, "_UnionGenericAlias"):
        if isinstance(thing, typing._UnionGenericAlias):
            return {
                "Union": [json_default(arg) for arg in get_args(thing)]
            }
    if thing == Literal[...]:
        return {
            "Literal": thing.__args__
        }
    if isinstance(thing, type(None)):
        return "None"
    if isinstance(thing, typing._SpecialForm):
        return thing._name
    if isinstance(thing, typing._GenericAlias) or isinstance(thing, types.GenericAlias):
        return {
            "GenericAlias": [json_default(arg) for arg in get_args(thing)]
        }
    if isinstance(thing, str):
        return thing
    if isinstance(thing, list) or isinstance(thing, tuple) or isinstance(thing, set):
        return [json_default(item) for item in thing]
    if isinstance(thing, dict):
        return {json_default(key): json_default(value) for key, value in thing.items()}

    raise TypeError(f"object of type {type(thing).__name__} not serializable")


def json_dumps(thing):
    return json.dumps(
        thing,
        default=json_default,
        ensure_ascii=False,
        sort_keys=True,
        indent=None,
        separators=(',', ':'),
    )


def get_model(content, logger, func_hash):
    """
    Get the model from the content and the logger.
    Decide on model depending on the length of the content. if is finetunable, return model, true, otherwise return model, false
    Args:
        content (str): the content to be aligned
        logger (buffered logger): the logger
        func_hash (str): the function hash
    Returns:
        model (str): the model to be used
        finetunable (bool): whether the model is finetunable
    """
    num_tokens = approximate_token_count(content)
    finetune_limit = logger.finetune_token_limit
    finetune_model, teacher_models = logger.get_models(func_hash)
    if num_tokens < finetune_limit:
        return finetune_model, True
    else:
        # this is just for backwards compatibility currently
        if len(teacher_models) == 0 or isinstance(teacher_models[0], str):
            teacher_models = [("gpt-4", 7000),("gpt-4-32k", 31000)]
        
        for model, token_limit in teacher_models:
            if num_tokens < token_limit:
                return model, False
        raise ValueError("The input content and align statements combined are too long, please shorten it. The maximum currently allowed token limit is 32000")
    

def approximate_token_count(content):
    """
    Approximate the token count of input
    Number of tokens is word tokens (nr of words * 1.33) + nr of special characters (which are usually their own tokens)
    Args:
        content (str, bytes): the content to be approximated
    Returns:
        number_of_tokens (int): the number of tokens
    """
    common_special_characters = r"\/(){}[]<>|`~@#$%^&*+=-_:;\""
    # check if input type is string
    if isinstance(content, str):
        number_of_word_tokens = int(len(content.split(" "))*1.333)
        nr_of_special_characters = sum([content.count(char) for char in common_special_characters])
        return number_of_word_tokens + nr_of_special_characters
    # check if input is a byte string
    if isinstance(content, bytes):
        number_of_word_tokens = int(len(content.split(b" "))*1.333)
        nr_of_special_characters = sum([content.count(char.encode("utf-8")) for char in common_special_characters])
        return number_of_word_tokens + nr_of_special_characters


def _deep_tuple(obj):
    """
    Convert a list or dict to a tuple recursively to allow for hashing and becoming a key for mock_behaviors
    :param obj:
    :return:
    """
    # transform pydantic objects into dicts
    if hasattr(obj, "__dict__"):
        obj = obj.__dict__
    if isinstance(obj, list) or isinstance(obj, tuple):
        return tuple(_deep_tuple(e) for e in obj)
    elif isinstance(obj, dict):
        return tuple((k, _deep_tuple(v)) for k, v in sorted(obj.items()))
    else:
        return obj


def get_key(args, kwargs) -> tuple:
    args_tuple = _deep_tuple(args)
    kwargs_tuple = _deep_tuple(kwargs)
    return args_tuple, kwargs_tuple


def prepare_object_for_saving(input_object):
    """
    Get a dictionary representation of the object
    """
    # check if list
    if isinstance(input_object, list):
        return [prepare_object_for_saving(item) for item in input_object]
    # check if tuple
    elif isinstance(input_object, tuple):
        return tuple([prepare_object_for_saving(item) for item in input_object])
    # check if dict
    elif isinstance(input_object, dict):
        return {key: prepare_object_for_saving(value) for key, value in input_object.items()}
    # check if pydantic object
    if hasattr(input_object, "__dict__"):
        attributes = input_object.__dict__
        for key, value in attributes.items():
            attributes[key] = prepare_object_for_saving(value)
        return attributes
        # 
    # check if datetime for custom logic
    elif isinstance(input_object, datetime.datetime):
        attrs = ['year', 'month', 'day', 'hour', 'minute', 'second', 'microsecond', 'tzinfo']
        attributes = {attr: getattr(input_object, attr, None) for attr in attrs}
        return attributes

    return input_object
def encode_int(n):
    # Define the character set for encoding
    charset = string.ascii_lowercase + string.digits + "_"
    return charset[n]

def decode_int(s):
    # Define the character set for encoding
    charset = string.ascii_lowercase + string.digits + "_"
    return charset.index(s)



from dotenv import load_dotenv
load_dotenv()
import tanuki
from pydantic import BaseModel
from typing import Literal


class Persona(BaseModel):
    email: str 
    name: str 
    company : str = None

@tanuki.patch
def classify_email(email: str) -> Literal["Real", "Fake"]:
    """
    Classify the email addresses as Fake or Real. The usual signs of an email being fake is the following:
    1) Using generic email addresses like yahoo, google, etc
    2) Misspellings in the email address
    3) Irregular name in email addresses
    """

@tanuki.align
def align_classify():
    assert classify_email("jeffrey.sieker@gmail.com") == "Fake"
    assert classify_email("jeffrey.sieker@apple.com") == "Real"
    assert classify_email("jon123121@apple.com") == "Fake"
    assert classify_email("jon@apple.com") == "Real"
    assert classify_email("jon.lorna@apple.com") == "Real"
    assert classify_email("jon.lorna@mircosoft.com") == "Fake"
    assert classify_email("jon.lorna@jklstarkka.com") == "Fake"
    assert classify_email("unicorn_rider123@yahoo.com") == "Fake"


@tanuki.patch
def extract_persona(email: str) -> Persona:
    """
    Using the email and email handler, extract the persona from the email
    The persona must have the email of the user,
    company (either the company name or None if generic Google, Yahoo etc email)
    name of the user to the best of the ability
    """

@tanuki.align
def align_extract():
    assert extract_persona("jeffrey.sieker@apple.com") == Persona(email="jeffrey.sieker@apple.com", name="Jeffrey Sieker", company="Apple")
    assert extract_persona("jon@amazon.com") == Persona(email="jon@apple.com", name="Jon", company="Amazon")
    assert extract_persona("jon.lorna@Lionmunch.com") == Persona(email="jon.lorna@apple.com", name="Jon Lorna", company="Lionmunch")
    assert extract_persona("jon.lorna@gmail.com") == Persona(email="jon.lorna@gmail.com", name="Jon Lorna")

def main(data_path, save_path):
    """
    Run through the workflow of the email cleaner
    First get data from the data_path
    Then call aligns for both MP functions
    Then classify emails and if real, extract personas
    Finally save personas to a excel file

    Args:
        data_path (str): the path to the data
        save_path (str): the path to save the personas to
    """
    # get data
    with open(data_path, "r") as f:
        emails = f.readlines()
        emails = [email.strip() for email in emails]

    # aligns
    align_classify()
    align_extract()

    personas = []
    # classify and extract
    for email in emails:
        output = classify_email(email)
        print(f"Checked {email} and classified as {output}")
        if output == "Real":
            personas.append(extract_persona(email))
    # save to excel
    import pandas as pd
    df = pd.DataFrame([persona.dict() for persona in personas])
    df.to_excel(save_path)

if __name__ == '__main__':
    data_path = r"examples\email_cleaner\data\test_emails.txt"
    save_path = r"examples\email_cleaner\data\personas.xlsx"
    main(data_path, save_path)


from dataclasses import dataclass

from pydantic import BaseModel

from tanuki.validator import Validator


def test_validate_output():
    print("test_validate_output")
    validator = Validator()
    assert validator.validate_output("1", int)
    assert validator.validate_output("1.0", float)
    assert validator.validate_output('"1"', str)
    assert validator.validate_output("true", bool)
    assert validator.validate_output("null", None)
    assert not validator.validate_output("1", float)
    assert not validator.validate_output("1.0", int)
    assert not validator.validate_output('"1"', int)
    #assert not validator.validate_output("true", int)
    assert not validator.validate_output("null", int)
    assert not validator.validate_output("1", str)
    assert not validator.validate_output("1.0", str)
    assert not validator.validate_output("true", str)

def test_validate_output_dataclass():
    print("test_validate_output_object")
    input_str = '{"name": "John", "age": 20, "height": 1.8, "is_cool": true}'

    @dataclass
    class Person:
        name: str
        age: int
        height: float
        is_cool: bool

        def __eq__(self, other):
            return self.dict() == other.dict()

        def __hash__(self):
            return hash(str(self.__dict__))

    validator = Validator()
    assert validator.validate_output(input_str, Person)

def test_validate_output_pydantic():
    print("test_validate_output_pydantic")
    class PersonPydantic(BaseModel):
        name: str
        age: int
        height: float
        is_cool: bool

    input_str = '{"name": "John", "age": 20, "height": 1.8, "is_cool": true}'
    validator = Validator()
    assert validator.validate_output(input_str, PersonPydantic)

if __name__ == "__main__":
    test_validate_output_dataclass()
    test_validate_output_pydantic()
    test_validate_output()


import openai
import os
from bs4 import BeautifulSoup
from dotenv import load_dotenv
from pydantic import BaseModel
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
import time
from typing import Optional

load_dotenv()

import tanuki


openai.api_key = os.getenv("OPENAI_API_KEY")


class AirBnb(BaseModel):
    city: str
    state: str
    dates: str
    price: float
    stars: float


@tanuki.patch
def extract_airbnb(content: str) -> Optional[AirBnb]:
    """
    Examine the content string and extract the airbnb details for the city, state,
    dates available, nightly price, and stars rating.
    """


@tanuki.align
def align_extract_airbnb() -> None:
    print("Aligning...")
    airbnb1 = "Caroga Lake, New YorkRoyal Mountain Ski ResortDec 3 – 8$200\xa0night$200 per night4.99"
    assert extract_airbnb(airbnb1) == AirBnb(
        city="Caroga Lake",
        state="New York",
        dates="Dec 3 - 8",
        price=200.0,
        stars=4.99,
    )


def selenium_driver() -> str:
    """Use selenium to scrape the airbnb url and return the page source."""

    # configure webdriver
    options = Options()
    # options.add_argument('--headless')  # Enable headless mode
    # options.add_argument('--disable-gpu')  # Disable GPU acceleration

    # launch driver for the page
    driver = webdriver.Chrome(options=options)
    driver.get("https://www.airbnb.com/?tab_id=home_tab&refinement_paths%5B%5D=%2Fhomes&search_mode=flex_destinations_search&flexible_trip_lengths%5B%5D=one_week&location_search=MIN_MAP_BOUNDS&monthly_start_date=2023-12-01&monthly_length=3&price_filter_input_type=0&channel=EXPLORE&search_type=category_change&price_filter_num_nights=5&category_tag=Tag%3A5366")
    time.sleep(3)

    # refresh the page to remove the dialog modal
    driver.refresh()
    time.sleep(3)

    # Scroll halfway down page to get rest of listings to load
    scroll_position = driver.execute_script("return (document.body.scrollHeight - window.innerHeight) * 0.4;")
    driver.execute_script(f"window.scrollTo(0, {scroll_position});")
    time.sleep(3)

    # extract the page source and return
    page_source = driver.page_source
    driver.quit()
    return page_source


if __name__ == '__main__':

    # Align the function
    align_extract_airbnb()

    # Selenium driver to scrape the url and extract the airbnb information
    page_source = selenium_driver()

    # Beautiful Soup to parse the page source
    soup = BeautifulSoup(page_source, 'html.parser')
    entities = soup.find_all('div', class_="dir dir-ltr")

    # Remove entries that are not airbnb listings
    contents = [entity.text for entity in entities if entity.text != ""]
    contents = [c for c in contents if "$" in c]
    print(contents)

    # Tanuki to extract the airbnb information
    print("Tanuki Time!")
    airbnbs = []
    for content in contents[1:3]:
        airbnbs.append(extract_airbnb(content))
    print(airbnbs)



import os
from typing import Literal, Optional

import openai
from dotenv import load_dotenv

import tanuki

load_dotenv()

openai.api_key = os.getenv("OPENAI_API_KEY")

@tanuki.patch
def classify_sentiment_2(input: str, input_2: str) -> Optional[Literal['Good', 'Bad']]:
    """
    Determine if the inputs are positive or negative sentiment, or None
    """


@tanuki.patch
def classify_sentiment(input: str) -> Optional[Literal['Good', 'Bad']]:
    """
    Determine if the input is positive or negative sentiment
    """


@tanuki.align
def align_classify_sentiment():
    """We can test the function as normal using Pytest or Unittest"""

    i_love_you = "I love you"
    print(classify_sentiment_2(i_love_you, "I love woo"))
    assert classify_sentiment_2(i_love_you, "I love woo") == 'Good'

    print(classify_sentiment("I love you"))
    assert classify_sentiment("I love you") == 'Good'

    assert classify_sentiment("I hate you") == 'Bad'
    #assert classify_sentiment("I hate you") != 'Good'
    assert not classify_sentiment("Wednesdays are in the middle of the week")


if __name__ == '__main__':
    align_classify_sentiment()



from dotenv import load_dotenv
load_dotenv()
import tanuki
import numpy as np


@tanuki.patch
def score_sentiment(input: str) -> Embedding[np.ndarray]:
    """
    Scores the input between 0-10
    """

@tanuki.align
def align_embed_sentiment() -> None:
    # We push these embeddings apart by declaring them to be different with the '!=' operator
    assert score_sentiment("I love this movie") != score_sentiment("I hate this movie")

    # And push these embeddings together by declaring equality
    assert score_sentiment("I love this movie") == score_sentiment("I love this film")
    assert score_sentiment("I love this movie") == score_sentiment("I loved watching the movie")

# This is a normal test that can be invoked
def test_score_sentiment():
    """We can test the function as normal using Pytest or Unittest"""
    align_embed_sentiment()
    print(score_sentiment("I like you"))
    print(score_sentiment("I do not like you"))
    print(score_sentiment("I am neutral"))


if __name__ == "__main__":
    test_score_sentiment()


from typing import Literal

from tanuki.bloom_filter import BloomFilter
from tanuki.persistence.filter.bloom_filtered_persistence import BloomFilteredPersistence
from tanuki.persistence.persistence_layer import FileSystemPersistence, RedisPersistence, S3Persistence
from tanuki.persistence.persistence_layer_interface import IPersistenceLayer

PersistenceType = Literal["filesystem", "redis", "s3"]

class PersistenceFactory:
    @staticmethod
    def create_persistence(type: PersistenceType, bloom_filter: BloomFilter) -> IPersistenceLayer:
        if type == "filesystem":
            return BloomFilteredPersistence(FileSystemPersistence(), bloom_filter)
        elif type == "redis":
            return BloomFilteredPersistence(RedisPersistence(), bloom_filter)
        elif type == "s3":
            return BloomFilteredPersistence(S3Persistence(), bloom_filter)
        else:
            raise ValueError("Unknown persistence type")



import datetime
import os
from contextlib import asynccontextmanager

import openai
from dotenv import load_dotenv
from fastapi import FastAPI
from typing import List

from pydantic import BaseModel

import tanuki
from todo_item import TodoItem

load_dotenv()
openai.api_key = os.getenv("OPENAI_API_KEY")


class Query(BaseModel):
    input: str


@tanuki.align
async def define_behavior():
    """
    We define 2 input/output pairs for the LLM to learn from.
    """

    assert create_todolist_item("I would like to go to the store and buy some milk") \
           == TodoItem(goal="Go to the store and buy some milk",
                       people=["Me"])

    assert create_todolist_item("I need to go and visit Jeff at 3pm tomorrow") \
           == TodoItem(goal="Go and visit Jeff",
                       people=["Me"],
                       deadline=datetime.datetime(2021, 1, 1, 15, 0))


@asynccontextmanager
async def lifespan(app: FastAPI):
    """
    This function allows us to run code before the FastAPI server starts (and after it stops)
    In this case, we use it to run the align statements on the `create_todolist_items` function.
    :param app:
    :return:
    """

    await define_behavior()
    yield

app = FastAPI(lifespan=lifespan)


@tanuki.patch
def create_todolist_items(input: str) -> List[TodoItem]:
    """
    Converts the input string into a list of TodoItem objects
    :param input: The user-supplied text of things they have to do
    :return: A list of TodoItem objects
    """


@tanuki.patch
def create_todolist_item(input: str) -> TodoItem:
    """
    Converts the input string into a TodoItem object
    :param input: The user-supplied text of things they have to do
    :return: TodoItem object
    """


@app.post("/create_todolist_items/")
async def create_todolist_items_route(input: Query):
    return create_todolist_items(input.input)


@app.post("/create_todolist_item/")
async def create_todolist_item_route(input: Query):
    return create_todolist_item(input.input)


if __name__ == "__main__":
    import uvicorn
    uvicorn.run(app, host="0.0.0.0", port=8000)


from dotenv import load_dotenv
load_dotenv()
import tanuki
from pydantic import BaseModel
from typing import Literal

class Response(BaseModel):
    """
    Response object, where the response attribute is the response sent to the customer
    requires_ticket is a boolean indicating whether the incoming tweet was a question or a direct issue
    """
    requires_ticket: bool 
    response: str 

class SupportTicket(BaseModel):
    """
    Support ticket object, where the name attribute is customers name,
    issue is a brief description of the issue customer had
    urgency conveys how urgently the team should respond to the issue
    """
    name: str
    issue: str
    urgency: Literal["low", "medium", "high"]

@tanuki.patch
def classify_and_respond(tweet_text: str) -> Response:
    """
    Respond to the customer support tweet text empathetically and nicely. 
    Convey that you care about the issue and if the problem was a direct issue that the support team should fix or a question, the team will respond to it. 
    """

@tanuki.align
def align_respond():
    input_tweet_1 = "Laia Johnson: I really like the new shovel but the handle broke after 2 days of use. Can I get a replacement?"
    assert classify_and_respond(input_tweet_1) == Response(
                                                            requires_ticket=True, 
                                                            response="Hi, we are sorry to hear that. We will get back to you with a replacement as soon as possible, can you send us your order nr?"
                                                            )
    input_tweet_2 = "Keira Townsend: I hate the new design of the iphone. It is so ugly. I am switching to Samsung"
    assert classify_and_respond(input_tweet_2) == Response(
                                                            requires_ticket=False, 
                                                            response="Hi, we are sorry to hear that. We will take this into consideration and let the product team know of the feedback"
                                                            )
    input_tweet_3 = "Thomas Bell: @Amazonsupport. I have a question about ordering, do you deliver to Finland?"
    assert classify_and_respond(input_tweet_3) == Response(
                                                            requires_ticket=True, 
                                                            response="Hi, thanks for reaching out. The question will be sent to our support team and they will get back to you as soon as possible"
                                                            )
    input_tweet_4 = "Jillian Murphy: Just bought the new goodybox and so far I'm loving it!"
    assert classify_and_respond(input_tweet_4) == Response(
                                                            requires_ticket=False, 
                                                            response="Hi, thanks for reaching out. We are happy to hear that you are enjoying the product"
                                                            )

@tanuki.patch
def create_support_ticket(tweet_text: str) -> SupportTicket:
    """
    Using the tweet text create a support ticket for saving to the internal database
    """

@tanuki.align
def align_supportticket():
    input_tweet_1 = "Laia Johnson: I really like the new shovel but the handle broke after 2 days of use. Can I get a replacement?"
    assert create_support_ticket(input_tweet_1) == SupportTicket(
                                                                name = "Laia Johnson", 
                                                                issue="Needs a replacement product because the handle broke", 
                                                                urgency = "high"
                                                                )
    input_tweet_2 = "Thomas Bell: @Amazonsupport. I have a question about ordering, do you deliver to Finland?"
    assert create_support_ticket(input_tweet_2) == SupportTicket(
                                                                name="Thomas Bell", 
                                                                issue="Answer whether we deliver to Finland", 
                                                                urgency="low"
                                                                )
    input_tweet_3 = "Jillian Murphy: Just bought the new goodybox and so far I'm loving it! The cream package was slightly damaged however, would need that to be replaced"
    assert classify_and_respond(input_tweet_3) == SupportTicket(
                                                                name="Jillian Murphy", 
                                                                issue="Needs a new cream as package was slightly damaged", 
                                                                urgency="medium"
                                                                )

def main():
    """
    Run through the workflow of twitter support bot.
    Example usecase uses 3 tweets, where first two require a support ticket and final one does not 
    """
    # start with calling aligns
    align_respond()
    align_supportticket()

    input_tweet_1 = "Jack Bell: Bro @Argos why did my order not arrive? I ordered it 2 weeks ago. Horrible service"
    response = classify_and_respond(input_tweet_1)
    # requires_ticket=True 
    # response="Hi Jack, we're really sorry to hear about this. We'll look into it right away and get back to you as soon as possible. Could you please provide us with your order number?"
    
    if response.requires_ticket:
        ticket = create_support_ticket(input_tweet_1)
        # name='Jack Bell' 
        # issue='Order did not arrive' 
        # urgency='high'
    
    input_tweet_2 = "Casey Montgomery: @Argos The delivery time was 3 weeks but was promised 1. Not a fan. "
    response = classify_and_respond(input_tweet_2)
    # requires_ticket=True 
    # response='Hi, we are sorry to hear about the delay. We will look into this issue and get back to you as soon as possible. Can you provide us with your order number?
    if response.requires_ticket:
        ticket = create_support_ticket(input_tweet_2)
        # name='Casey Montgomery' 
        # issue='Complaint about delivery time' 
        # urgency='medium'
    
    input_tweet_3 = "Jacks Parrow: @Argos The new logo looks quite ugly, wonder why they changed it"
    response = classify_and_respond(input_tweet_3)
    # requires_ticket=False 
    # response='Hi, we are sorry to hear that. We will take this into consideration and let the product team know of the feedback'

    if response.requires_ticket:
        ticket = create_support_ticket(input_tweet_3)
        # No ticket

if __name__ == '__main__':
    main()


import logging
import os
from datetime import datetime
from typing import List
from xml.etree import ElementTree as ET

import openai
import requests
from dotenv import load_dotenv
from pydantic import BaseModel, ConfigDict, Field
from requests_html import HTMLSession
from selenium import webdriver
from webdriver_manager.chrome import ChromeDriverManager
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from bs4 import BeautifulSoup

from examples.google_news_scraper.utils import send_email
import tanuki

load_dotenv()

openai.api_key = os.getenv("OPENAI_API_KEY")

# Define Pydantic model of an article summary
class ArticleSummary(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    impact: int = Field(..., ge=0, le=10)
    sentiment: float = Field(..., ge=-1.0, le=1.0)
    date: datetime.date
    companies_involved: List[str]
    people_involved: List[str]
    summary: str

def configure_selenium_user_agent():
    options = Options()
    options.add_argument("--headless")
    options.add_argument("--no-sandbox")
    options.add_argument("--disable-gpu")
    options.add_argument("--window-size=1920,1080")
    options.add_argument("--disable-extensions")
    options.add_argument("--disable-dev-shm-usage")
    options.add_argument("--lang=en-US,en")
    options.add_argument("user-agent=Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36")
    return options


def get_absolute_redirect_url_from_google_rss(url):
    """
    This function uses requests-html to get the absolute URL of a give link in a Google RSS feed.
    :param url:
    :return:
    """
    session = HTMLSession()
    response = session.get(url, allow_redirects=True)
    try:
        response.html.render(timeout=20)
    except Exception as e:
        print(f"An error occurred: {e}")
        return None
    finally:
        session.close()

    absolute_links = list(response.html.absolute_links)
    redirected_url = absolute_links[0] if absolute_links else None
    return redirected_url


def parse_article_with_selenium(url: str) -> str:
    """
    This function uses Selenium to extract the text of an article from a given URL.
    """

    options = configure_selenium_user_agent()
    service = Service(ChromeDriverManager().install())
    driver = webdriver.Chrome(service=service, options=options)

    try:
        driver.get(url)
        WebDriverWait(driver, 10).until(EC.presence_of_element_located((By.TAG_NAME, "body")))
        soup = BeautifulSoup(driver.page_source, 'html.parser')
        # Perform your parsing with BeautifulSoup here

        # For example, to get text without tags:
        article_text = soup.get_text(separator=' ', strip=True)
        return article_text
    finally:
        driver.quit()


def scrape_google_news(search_term: str, recipient: str, max=5):
    """
    This function scrapes Google News for articles about a given search term.
    :param search_term:
    :param recipient:
    :param max:
    :return:
    """
    # RSS feed URL with the search term
    url = f"https://news.google.com/rss/search?q={search_term}&hl=en-US&gl=US&ceid=US:en"
    response = requests.get(url)
    root = ET.fromstring(response.content)
    items = root.findall('./channel/item')

    logging.info(f"Found {len(items)} articles for {search_term}")

    relevant_articles = []
    for item in items[:max]:
        # Extract the link to the full article
        link = item.find('link').text

        final_url = get_absolute_redirect_url_from_google_rss(link)
        if final_url:
            article_content = parse_article_with_selenium(final_url)
            try:
                article_summary = analyze_article(article_content, search_term)
                logging.info(article_summary)

                # Check if the article is relevant based on impact and sentiment
                if article_summary.impact > 5 and article_summary.sentiment < 0:
                    relevant_articles.append(article_summary)
            except Exception as e:
                logging.warning(f"An error occurred: {final_url=} {e}")
                continue

    return relevant_articles


def email_if_relevant(relevant_articles: List[ArticleSummary], search_term: str, recipient: str):
    """
    This function sends an email if relevant articles were found relating to the search term.
    :param relevant_articles: A list of relevant articles extracted from a website.
    :param search_term:
    :param recipient:
    :return:
    """
    if relevant_articles:
        subject = f"Summary of Important Articles about {search_term}"
        body = "The following articles about {search_term} have high impact and negative sentiment:\n\n"
        for summary in relevant_articles:
            body += f"- {summary.summary} (Impact: {summary.impact}, Sentiment: {summary.sentiment})\n"

        send_email(subject, body, recipient)


@tanuki.patch
def analyze_article(html_content: str, subject: str) -> ArticleSummary:
    """
    Analyzes the article's HTML content and extracts information relevant to the subject.
    """


@tanuki.align
def align_analyze_article():

    html_content = "<head></head><body><p>Nvidia has made the terrible decision to buy ARM for $40b on 8th November. This promises to "\
                   "be an extremely important decision for the industry, even though it creates a monopoly.</p></body> "
    assert analyze_article(html_content, "nvidia") == ArticleSummary(
        impact=10,
        sentiment=-0.9,
        date=datetime.date(2023, 11, 8),
        companies_involved=["Nvidia", "ARM"],
        people_involved=[],
        summary="Nvidia is acquiring ARM for $40 billion, which will have a huge impact on the semiconductor industry.",
    )




# Example usage:
if __name__ == '__main__':
    align_analyze_article()

    recipient = 'dummy@example.com'
    search_term = 'nvidia'
    relevant_articles = scrape_google_news(search_term, recipient)
    email_if_relevant(relevant_articles, search_term, recipient)



class IPersistenceLayer:
    def store(self, key, data):
        pass

    def retrieve(self, key):
        pass

    def update(self, key, data):
        pass

    def exists(self, key):
        pass



import openai
import os
from dotenv import load_dotenv
from pydantic import BaseModel
from typing import List, Optional

load_dotenv()

import tanuki
from utils import scrape_url


openai.api_key = os.getenv("OPENAI_API_KEY")


class Cocktail(BaseModel):
    name: str
    ingredients: List[str] = []
    instructions: str
    similar: List[str] = []


@tanuki.patch
def extract_cocktail(content: str) -> Optional[Cocktail]:
    """
    Examine the content string and extract the cocktail details for the ingredients, instructions, and similar cocktails.
    """


@tanuki.align
def align_extract_cocktail() -> None:
    print("Aligning...")
    cocktail = """Black Rose | Kindred Cocktails\n\n\n\n\n\n      Skip to main content\n    \n\n\n\n\n\nKindred Cocktails\n\n\nToggle navigation\n\n\n\n\n\n\n\n\nMain navigation\n\n\nHome\n\n\nCocktails\n\n\nNew\n\n\nInfo \n\n\nStyle guidelines\n\n\nIngredients\n\n\n\n\n\nMeasurement units\n\n\nHistoric Cocktail Books\n\n\nRecommended Brands\n\n\nAmari & Friends\n\n\nArticles & Reviews\n\n\n\n\n\nAbout us\n\n\nLearn More\n\n\nFAQ\n\n\nTerms of Use\n\n\nContact us\n\n\n\n\nYou \n\n\nLog in\n\n\nSign Up\n\n\nReset your password\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\nHome\n\n\nCocktails\n\n\n                  Black Rose\n              \n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\nCopy\n\n\n\n\nBlack Rose\n \n\n\n\n\n\n\n\n\n\n2 oz Bourbon\n\n1 ds Grenadine\n\n2 ds Peychaud's Bitters\n\n1  Lemon peel (flamed, for garnish)\n\n\n\nInstructions\nFill an old-fashioned glass three-quarters full with ice.  Add the bourbon, grenadine, and bitters, and stir.  Garnish with the lemon peel.\n\n\n\n\n\n\nCocktail summary\n\n\n\nPosted by\nThe Boston Shaker\n on \n4/12/2011\n\n\n\n\nIs of\nunknown authenticity\n\n\nReference\nDale Degroff, The Essential Cocktail, p48\n\n\n\nCurator\nNot yet rated\n\n\nAverage\n3.5 stars (6 ratings)\n\n\n\nYieldsDrink\n\n\nScale\n\n\nBourbon, Peychaud's Bitters, Grenadine, Lemon peel\nPT5M\nPT0M\nCocktail\nCocktail\n1\ncraft, alcoholic\n3.66667\n6\n\n\n\n\n\n\n\n\n\n\nCocktail Book\n\nLog in or sign up to start building your Cocktail Book.\n\n\n\n\nFrom other usersWith a modest grenadine dash, this drink didn't do much for me, but adding a bit more won me over.\nSimilar cocktailsNew Orleans Cocktail — Bourbon, Peychaud's Bitters, Orange Curaçao, Lemon peelOld Fashioned — Bourbon, Bitters, Sugar, Lemon peelBattle of New Orleans — Bourbon, Peychaud's Bitters, Absinthe, Orange bitters, Simple syrupImproved Whiskey Cocktail — Bourbon, Bitters, Maraschino Liqueur, Absinthe, Simple syrup, Lemon peelDerby Cocktail — Bourbon, Bénédictine, BittersMother-In-Law — Bourbon, Orange Curaçao, Maraschino Liqueur, Peychaud's Bitters, Bitters, Torani Amer, Simple syrupMint Julep — Bourbon, Rich demerara syrup 2:1, MintThe Journey — Bourbon, Mezcal, Hazelnut liqueurBenton's Old Fashioned — Bourbon, Bitters, Grade B maple syrup, Orange peelFancy Mint Julep — Bourbon, Simple syrup, Mint, Fine sugar\n\nComments\n\n\n\n\n\nLog in or register to post comments\n\n\n\n\n\n\n\n\n© 2010-2023 Dan Chadwick. Kindred Cocktails™ is a trademark of Dan Chadwick."""
    assert extract_cocktail(cocktail) == Cocktail(
        name="Black Rose",
        ingredients=["2 oz Bourbon", "1 ds Grenadine", "2 ds Peychaud's Bitters", "1  Lemon peel (flamed, for garnish)"],
        instructions="Fill an old-fashioned glass three-quarters full with ice.  Add the bourbon, grenadine, and bitters, and stir.  Garnish with the lemon peel.",
        similar=["New Orleans Cocktail", "Old Fashioned", "Battle of New Orleans", "Improved Whiskey Cocktail", "Derby Cocktail", "Mother-In-Law", "Mint Julep", "The Journey", "Benton's Old Fashioned", "Fancy Mint Julep"],
    )


if __name__ == '__main__':

    # Align the function
    align_extract_cocktail()

    # Web scrape the url and extract the cocktail information
    url = "https://kindredcocktails.com/cocktail/old-fashioned"
    # url = "https://kindredcocktails.com/cocktail/journey"
    contents = scrape_url(url=url)
    print(contents)

    # Process the cocktail block using Tanuki
    cocktail = extract_cocktail(contents[0])
    print(cocktail)



from dataclasses import dataclass
from typing import List, Tuple, Set, Dict, Mapping, MutableMapping, OrderedDict, ChainMap, Counter, DefaultDict, Deque, \
    MutableSequence, Sequence, Union, Literal

from tanuki.validator import Validator


def test_validate_base_types():
    print("test_validate_base_types")
    validator = Validator()
    assert validator.check_type(1, int)
    assert validator.check_type(1.0, float)
    assert validator.check_type("1", str)
    assert validator.check_type(True, bool)
    assert validator.check_type(None, None)
    assert not validator.check_type(1, float)
    assert not validator.check_type(1.0, int)
    assert not validator.check_type("1", int)
    # assert not validator.validate_value(True, int)
    assert not validator.check_type(None, int)
    assert not validator.check_type(1, str)
    assert not validator.check_type(1.0, str)
    assert not validator.check_type(True, str)
    assert not validator.check_type(None, str)
    assert not validator.check_type(1, bool)
    assert not validator.check_type(1.0, bool)
    assert not validator.check_type("1", bool)
    assert not validator.check_type(None, bool)
    assert not validator.check_type(1, None)
    assert not validator.check_type(1.0, None)
    assert not validator.check_type("1", None)
    assert not validator.check_type(True, None)


def test_validate_complex_types():
    print("test_validate_complex_types")
    validator = Validator()
    assert validator.check_type([1, 2, 3], list)
    assert validator.check_type([1, 2, 3], list[int])
    assert not validator.check_type([1, 2, 3], list[str])
    assert validator.check_type((1, 2, 3), tuple)
    assert validator.check_type({1, 2, 3}, set)
    assert validator.check_type({"a": 1, "b": 2, "c": 3}, dict)
    assert not validator.check_type([1, 2, 3], tuple)
    assert not validator.check_type((1, 2, 3), list)
    assert not validator.check_type({1, 2, 3}, dict)
    assert not validator.check_type({"a": 1, "b": 2, "c": 3}, set)
    assert not validator.check_type([1, 2, 3], dict)
    assert not validator.check_type((1, 2, 3), set)
    assert not validator.check_type({1, 2, 3}, list)
    assert not validator.check_type({"a": 1, "b": 2, "c": 3}, tuple)
    assert not validator.check_type([1, 2, 3], None)
    assert not validator.check_type((1, 2, 3), None)


def test_validate_type_annotations():
    print("test_validate_type_annotations")
    validator = Validator()
    assert validator.check_type([1, 2], List[int])
    assert validator.check_type([1, 2], List)
    assert not validator.check_type([1, 2], List[str])

    assert validator.check_type((1, 2), Tuple[int, int])
    assert validator.check_type((1, 2), Tuple)
    assert not validator.check_type((1, 2), Tuple[str, str])

    assert validator.check_type({1, 2}, Set[int])
    assert validator.check_type({1, 2}, Set)
    assert not validator.check_type({1, 2}, Set[str])

    assert validator.check_type({"a": 1, "b": 2}, Dict[str, int])
    assert validator.check_type({"a": 1, "b": 2}, Dict)
    assert not validator.check_type({"a": 1, "b": 2}, Dict[str, str])


def test_validate_collection_dict_types():
    print("test_validate_collection_types")
    validator = Validator()
    assert validator.check_type({"a": 1, "b": 2}, OrderedDict[str, int])
    assert validator.check_type({"a": 1, "b": 2}, OrderedDict)

    assert validator.check_type({"a": 1, "b": 2}, MutableMapping[str, int])
    assert validator.check_type({"a": 1, "b": 2}, MutableMapping)

    assert validator.check_type({"a": 1, "b": 2}, Mapping[str, int])
    assert validator.check_type({"a": 1, "b": 2}, Mapping)

    assert validator.check_type({"a": 1, "b": 2}, ChainMap[str, int])
    assert validator.check_type({"a": 1, "b": 2}, ChainMap)

    assert validator.check_type({"a": 1, "b": 2}, Counter[str])
    assert validator.check_type({"a": 1, "b": 2}, Counter)

    assert validator.check_type({"a": 1, "b": 2}, DefaultDict[str, int])
    assert validator.check_type({"a": 1, "b": 2}, DefaultDict)


def test_validate_collection_list_types():
    print("test_validate_collection_list_types")
    validator = Validator()
    assert validator.check_type([1, 2], Deque[int])
    assert validator.check_type([1, 2], Deque)

    assert validator.check_type([1, 2], MutableSequence[int])
    assert validator.check_type([1, 2], MutableSequence)

    assert validator.check_type([1, 2], Sequence[int])
    assert validator.check_type([1, 2], Sequence)

    assert not validator.check_type([1, 2], Deque[str])
    assert not validator.check_type([1, 2], MutableSequence[str])
    assert not validator.check_type([1, 2], Sequence[str])

    assert validator.check_type([1, '2'], Sequence[Union[int, str]])


def test_validate_literal_types():
    print("test_validate_literal_types")
    validator = Validator()
    assert validator.check_type(1, Literal[1])
    assert validator.check_type(1, Literal[1, 2])
    assert validator.check_type(1, Literal[1, 2, 3])
    assert validator.check_type(1, Literal[1, 2, 3, 4])
    assert validator.check_type('Good', Literal['Good', 'Bad'])

    assert validator.check_type(1, Literal[1.0, 2.0, 3.0, 4.0])
    assert validator.check_type(1, Literal[1.0, 2.0, 3.0])
    assert validator.check_type(1, Literal[1.0, 2.0])
    assert validator.check_type(1, Literal[1.0])

    assert not validator.check_type(1, Literal[2])
    assert not validator.check_type(1, Literal[2, 3])

    assert not validator.check_type(1, Literal[2.0])
    assert not validator.check_type(1, Literal['Good', 'Bad'])


def test_validate_dataclasses():
    print("test_validate_dataclasses")
    validator = Validator()

    @dataclass
    class Person:
        name: str
        age: int
        height: float
        is_cool: bool

    person = Person(name='John', age=20, height=1.8, is_cool=True)
    person = {'name': 'John', 'age': 20, 'height': 1.8, 'is_cool': True}
    assert validator.check_type(person, Person)
    assert not validator.check_type(person, str)


if __name__ == "__main__":
    test_validate_dataclasses()
    test_validate_literal_types()
    test_validate_collection_list_types()
    test_validate_collection_dict_types()
    test_validate_type_annotations()
    test_validate_complex_types()
    test_validate_base_types()



from tanuki.validator import Validator


def test_validate_base_type():
    print("test_validate_base_type")
    validator = Validator()
    assert validator.validate_base_type(1, int)
    assert validator.validate_base_type(1.0, float)
    assert validator.validate_base_type("1", str)
    assert validator.validate_base_type(True, bool)
    assert validator.validate_base_type(None, None)
    assert not validator.validate_base_type(1, float)
    assert not validator.validate_base_type(1.0, int)
    assert not validator.validate_base_type("1", int)
    #assert not validator.validate_base_type(True, int)
    assert not validator.validate_base_type(None, int)
    assert not validator.validate_base_type(1, str)
    assert not validator.validate_base_type(1.0, str)
    assert not validator.validate_base_type(True, str)
    assert not validator.validate_base_type(None, str)
    assert not validator.validate_base_type(1, bool)
    assert not validator.validate_base_type(1.0, bool)
    assert not validator.validate_base_type("1", bool)
    assert not validator.validate_base_type(None, bool)
    assert not validator.validate_base_type(1, None)
    assert not validator.validate_base_type(1.0, None)
    assert not validator.validate_base_type("1", None)
    assert not validator.validate_base_type(True, None)

if __name__ == "__main__":
    test_validate_base_type()


import hashlib
import logging
import math

import numpy as np
from bitarray import bitarray

from tanuki.persistence.filter.bloom_interface import IBloomFilterPersistence


class BloomFilter:

    def __init__(self,
                 persistence: IBloomFilterPersistence,
                 size=None,
                 hash_count=None,
                 expected_number_of_elements=None,
                 false_positive_probability=None):

        if not persistence:
            raise ValueError("Persistence cannot be None, it must be an instance of IBloomFilterPersistence")

        if not size and not hash_count and not expected_number_of_elements and not false_positive_probability:
            raise ValueError("Must specify either (size, hash_count) or (expected_number_of_elements, false_positive_probability")

        if expected_number_of_elements and false_positive_probability:
            size, hash_count = BloomFilter.optimal_bloom_filter_params(expected_number_of_elements, false_positive_probability)

        if not size and not hash_count:
            raise ValueError("Size and hash_count not set. This should never happen.")

        self.size = size
        self.hash_count = hash_count
        self.bit_array, self.indices = self.init_bit_array(size)
        self.persistence = persistence

    def init_bit_array(self, size):
        _bit_array = bitarray(size)
        _bit_array.setall(0)
        _indices = np.zeros(size, dtype=np.int32)
        return _bit_array, _indices

    def hash_functions(self, string):
        # h1(x)
        hash1 = int(hashlib.sha256(string.encode('utf-8')).hexdigest(), 16)
        # h2(x)
        hash2 = int(hashlib.md5(string.encode('utf-8')).hexdigest(), 16)
        return hash1, hash2

    def lookup(self, string):
        hash1, hash2 = self.hash_functions(string)
        for seed in range(self.hash_count):
            index = (hash1 + seed * hash2) % self.size

            #print(f"Lookup: Seed={seed}, Digest={index}, BitValue={self.bit_array[index]}")
            if self.bit_array[index] == 0:
                return False
        return True

    def add(self, string):
        hash1, hash2 = self.hash_functions(string)
        for seed in range(self.hash_count):
            index = (hash1 + seed * hash2) % self.size
            self.bit_array[index] = 1
            #print(f"Add: Seed={seed}, Digest={index}, BitValue={self.bit_array[index]}")

    def save(self):
        self.persistence.save(self.bit_array)

    def load(self):
        self.bit_array = self.persistence.load()

        length_in_bytes = int(len(self.bit_array)/8)
        expected_length = math.ceil(self.size / 8)
        if length_in_bytes != expected_length:
            logging.warning("Bit array length does not match expected size, and so might be corrupted. Reinitializing.")
            self.bit_array, self.indices = self.init_bit_array(self.size)
            self.save()



    @staticmethod
    def optimal_bloom_filter_params(n, p):
        """
        Calculate the optimal bit array size (m) and number of hash functions (k)
        for a Bloom filter.

        n: expected number of items to be stored
        p: acceptable false positive probability

        Returns a tuple (m, k)
        """
        m = - (n * math.log(p)) / (math.log(2) ** 2)
        k = (m / n) * math.log(2)
        return int(math.ceil(m)), int(math.ceil(k))